"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import logging
import os
from test.testlib.testcase import BaseTestCase
from unittest.mock import patch

import cfnlint.config  # pylint: disable=E0401

LOGGER = logging.getLogger("cfnlint")


class TestArgsParser(BaseTestCase):
    """Test Parser Arguments"""

    def tearDown(self):
        """Setup"""
        for handler in LOGGER.handlers:
            LOGGER.removeHandler(handler)

    def test_create_parser(self):
        """Test success run"""

        config = cfnlint.config.CliArgs(
            ["-t", "test.yaml", "--ignore-bad-template", "--format", "quiet", "--debug"]
        )
        self.assertEqual(config.cli_args.templates, [])
        self.assertEqual(config.cli_args.template_alt, ["test.yaml"])
        self.assertEqual(config.cli_args.ignore_bad_template, True)
        self.assertEqual(config.cli_args.format, "quiet")
        self.assertEqual(config.cli_args.debug, True)

    def test_create_parser_default_param(self):
        """Test success run"""

        config = cfnlint.config.CliArgs(
            [
                "--regions",
                "us-east-1",
                "us-west-2",
                "--",
                "template1.yaml",
                "template2.yaml",
            ]
        )
        self.assertEqual(
            config.cli_args.templates, ["template1.yaml", "template2.yaml"]
        )
        self.assertEqual(config.cli_args.template_alt, [])
        self.assertEqual(config.cli_args.regions, ["us-east-1", "us-west-2"])

    def test_stdout(self):
        """Test success run"""

        config = cfnlint.config.CliArgs(["-t", "template1.yaml"])
        self.assertIsNone(config.cli_args.output_file)

    def test_output_file(self):
        """Test success run"""

        config = cfnlint.config.CliArgs(
            ["-t", "template1.yaml", "--output-file", "test_output.txt"]
        )
        self.assertEqual(config.cli_args.output_file, "test_output.txt")

    def test_force_update_specs(self):
        """Test success run"""

        config = cfnlint.config.CliArgs(["--update-specs", "--force"])
        self.assertEqual(config.cli_args.force, True)
        self.assertEqual(config.cli_args.update_specs, True)

    def test_create_parser_exend(self):
        """Test success run"""

        config = cfnlint.config.CliArgs(
            ["-t", "template1.yaml", "-t", "template2.yaml"]
        )
        self.assertEqual(config.cli_args.templates, [])
        self.assertEqual(
            config.cli_args.template_alt, ["template1.yaml", "template2.yaml"]
        )

    def test_create_parser_config_file_regions(self):
        """Test success run"""

        config = cfnlint.config.CliArgs(
            [
                "--regions",
                "us-west-1",
                "--include-checks",
                "I1234",
                "--",
                "template1.yaml",
            ]
        )
        self.assertEqual(config.cli_args.templates, ["template1.yaml"])
        self.assertEqual(config.cli_args.include_checks, ["I1234"])
        self.assertEqual(config.cli_args.regions, ["us-west-1"])

    def test_create_parser_config_file(self):
        """Test success run"""

        config = cfnlint.config.CliArgs(
            ["--mandatory-checks", "I1234", "--", "template1.yaml"]
        )
        self.assertEqual(config.cli_args.templates, ["template1.yaml"])
        self.assertEqual(config.cli_args.mandatory_checks, ["I1234"])

    def test_create_parser_rule_configuration(self):
        """Test success run"""

        config = cfnlint.config.CliArgs(
            ["-x", "E3012:strict=true", "-x", "E3012:key=value,E3001:key=value"]
        )
        self.assertEqual(
            config.cli_args.configure_rules,
            {"E3012": {"key": "value", "strict": "true"}, "E3001": {"key": "value"}},
        )

    def test_exit_code_parameter(self):
        """Test values of exit code"""

        for param in ["informational", "warning", "error"]:
            with self.subTest():
                config = cfnlint.config.CliArgs(["--non-zero-exit-code", param])
                self.assertEqual(config.cli_args.non_zero_exit_code, param)

    def test_exit_code_parameter_error(self):
        """Test result when bad value provided"""
        with open(os.devnull, "w") as devnull:
            with patch("sys.stderr", devnull):
                with self.assertRaises(SystemExit):
                    cfnlint.config.CliArgs(["--non-zero-exit-code", "bad"])



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from test.unit.rules import BaseRuleTestCase

from cfnlint.context import Context, Path
from cfnlint.jsonschema import CfnTemplateValidator
from cfnlint.rules.parameters.Enum import Enum


class TestAllowedValue(BaseRuleTestCase):
    """Test Allowed Value Parameter Configuration"""

    def setUp(self):
        """Setup"""
        self.rule = Enum()

    def test_validate(self):
        validator = CfnTemplateValidator(
            schema={"type": "string"},
            context=Context(
                "us-east-1",
                {},
                path=Path(
                    path=deque(["Property", "Ref"]),
                    value_path=deque(["Parameters", "MyParameter", "Default"]),
                ),
            ),
        )

        errs = list(self.rule.enum(validator, "1", ["A", "B"], {}))
        self.assertEqual(len(errs), 1)
        self.assertEqual(len(errs), 1)
        for err in errs:
            self.assertEqual(err.message, "['A', 'B'] is not one of '1'")
            self.assertEqual(err.rule, self.rule)
            self.assertEqual(
                err.path_override, deque(["Parameters", "MyParameter", "Default"])
            )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import regex as re

from cfnlint._typing import RuleMatches
from cfnlint.rules import CloudFormationLintRule, RuleMatch
from cfnlint.template import Template


class HardCodedArnProperties(CloudFormationLintRule):
    """
    Checks Resources if ARNs use correctly placed Pseudo Parameters
    instead of hardcoded Partition, Region, and Account Number
    """

    id = "I3042"
    shortdesc = "ARNs should use correctly placed Pseudo Parameters"
    description = (
        "Checks Resources if ARNs use correctly placed Pseudo Parameters instead of"
        " hardcoded Partition, Region, and Account Number"
    )
    source_url = ""
    tags = ["resources"]
    regex = re.compile(
        r"arn:(\$\{[^:]*::[^:]*}|[^:]*):[^:]+:(\$\{[^:]*::[^:]*}|[^:]*):(\$\{[^:]*::[^:]*}|[^:]*)"
    )

    def __init__(self):
        """Init"""
        super().__init__()
        self.config_definition = {
            "partition": {
                "default": True,
                "type": "boolean",
            },
            "region": {
                "default": False,
                "type": "boolean",
            },
            "accountId": {
                "default": False,
                "type": "boolean",
            },
        }
        self.configure()

    def _match_values(self, cfnelem, path):
        """Recursively search for values matching the searchRegex"""
        values = []
        if isinstance(cfnelem, dict):
            for key in cfnelem:
                pathprop = path[:]
                pathprop.append(key)
                values.extend(self._match_values(cfnelem[key], pathprop))
        elif isinstance(cfnelem, list):
            for index, item in enumerate(cfnelem):
                pathprop = path[:]
                pathprop.append(index)
                values.extend(self._match_values(item, pathprop))
        else:
            # Leaf node
            if isinstance(cfnelem, str):  # and re.match(searchRegex, cfnelem):
                for variable in re.findall(self.regex, cfnelem):
                    if "Fn::Sub" in path:
                        values.append(path + [variable])

        return values

    def match_values(self, cfn):
        """
        Search for values in all parts of the templates that match the searchRegex
        """
        results = []
        results.extend(self._match_values(cfn.template.get("Resources", {}), []))
        # Globals are removed during a transform.  They need to be checked manually
        results.extend(self._match_values(cfn.template.get("Globals", {}), []))
        return results

    def match(self, cfn: Template) -> RuleMatches:
        matches: RuleMatches = []

        transforms = cfn.transform_pre["Transform"]
        transforms = transforms if isinstance(transforms, list) else [transforms]
        if "AWS::Serverless-2016-10-31" in cfn.transform_pre["Transform"]:
            return matches

        # Get a list of paths to every leaf node string containing at least one ${parameter}
        parameter_string_paths = self.match_values(cfn)
        # We want to search all of the paths to check if each one contains an 'Fn::Sub'
        for parameter_string_path in parameter_string_paths:
            path = ["Resources"] + parameter_string_path[:-1]
            candidate = parameter_string_path[-1]

            # ruff: noqa: E501
            # !Sub arn:${AWS::Partition}:iam::aws:policy/service-role/AWSLambdaBasicExecutionRole
            # is valid even with aws as the account #.  This handles empty string
            if self.config["partition"] and not re.match(
                r"^\$\{\w+}|\$\{AWS::Partition}|$", candidate[0]
            ):
                # or not re.match(r'^(\$\{\w+}|\$\{AWS::Region}|)$',candidate[1])
                # or not re.match(r'^\$\{\w+}|\$\{AWS::AccountId}|aws|$', candidate[2]):
                message = (
                    "ARN in Resource {0} contains hardcoded Partition in ARN or"
                    " incorrectly placed Pseudo Parameters"
                )
                matches.append(RuleMatch(path, message.format(path[1])))
            if self.config["region"] and not re.match(
                r"^(\$\{\w+}|\$\{AWS::Region}|)$", candidate[1]
            ):
                # or  or not re.match(r'^\$\{\w+}|\$\{AWS::AccountId}|aws|$', candidate[2]):
                message = (
                    "ARN in Resource {0} contains hardcoded Region in ARN or"
                    " incorrectly placed Pseudo Parameters"
                )
                matches.append(RuleMatch(path, message.format(path[1])))
            if self.config["accountId"] and not re.match(
                r"^\$\{\w+}|\$\{AWS::AccountId}|aws|$", candidate[2]
            ):
                message = (
                    "ARN in Resource {0} contains hardcoded AccountId in ARN or"
                    " incorrectly placed Pseudo Parameters"
                )
                matches.append(RuleMatch(path, message.format(path[1])))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.resources.ectwo.RouteTableAssociation import (
    RouteTableAssociation,  # pylint: disable=E0401
)


class TestPropertyRtAssociation(BaseRuleTestCase):
    """Test Route Table Association"""

    def setUp(self):
        """Setup"""
        super(TestPropertyRtAssociation, self).setUp()
        self.collection.register(RouteTableAssociation())

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_file_negative(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/properties_rt_association.yaml", 5
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule


class FindInMapResolved(CloudFormationLintRule):
    id = "W1034"
    shortdesc = "Validate the values that come from a Fn::FindInMap function"
    description = (
        "Resolve the Fn::FindInMap and then validate the values against the schema"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-findinmap.html"
    tags = ["functions", "findinmap"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.testlib.testcase import BaseTestCase

from cfnlint import ConfigMixIn, Template
from cfnlint.decode.mark import Mark
from cfnlint.decode.node import dict_node
from cfnlint.exceptions import DuplicateRuleError
from cfnlint.rules import CloudFormationLintRule, Match, RuleMatch, Rules
from cfnlint.rules.errors import RuleError


class Rule(CloudFormationLintRule):
    """Def Rule"""

    id = "EXXXX"
    shortdesc = "Test Rule"
    description = "Test Rule"
    source_url = "https://github.com/aws-cloudformation/cfn-lint/"
    tags = ["resources"]

    def __init__(self):
        """Init"""
        super(Rule, self).__init__()
        self.config_definition = {
            "testBoolean": {"default": True, "type": "boolean"},
            "testString": {"default": "default", "type": "string"},
            "testInteger": {"default": 1, "type": "integer"},
        }
        self.configure()

    def get_config(self):
        """Get the Config"""
        return self.config


class RuleFail(CloudFormationLintRule):
    """Def Rule"""

    id = "EYYYY"
    shortdesc = "Test Rule"
    description = "Test Rule"
    source_url = "https://github.com/aws-cloudformation/cfn-lint/"
    tags = ["resources"]

    def match(self, cfn):
        raise KeyError("Bad template")


class RuleChild(CloudFormationLintRule):
    id = "ECCCC"
    shortdesc = "Child Rule"
    description = "Child Rule"
    source_url = "https://github.com/aws-cloudformation/cfn-lint/"
    tags = ["resources"]

    def validate(self, cfn):
        return [RuleMatch([], "Child Rule", rule=self)]


class RuleParent(CloudFormationLintRule):
    id = "EPPPP"
    shortdesc = "Parent Rule"
    description = "Parent Rule"
    source_url = "https://github.com/aws-cloudformation/cfn-lint/"
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__()
        self.child_rules = {
            "ECCCC": RuleChild(),
        }

    def match(self, cfn):
        return [RuleMatch([], "Parent Rule")] + self.child_rules["ECCCC"].validate(cfn)


class TestRules(BaseTestCase):
    """Test CloudFormation Rules"""

    def test_rule_failure(self):
        rules = Rules()
        rules.extend([RuleFail(), RuleError()])

        cfn = Template("-", {}, regions=["us-east-1"])

        matches = list(rules.run("-", cfn, ConfigMixIn([])))
        self.assertListEqual(
            [
                Match(
                    linenumber=1,
                    linenumberend=1,
                    columnnumber=1,
                    columnnumberend=2,
                    filename="-",
                    rule=RuleError(),
                    message=(
                        "Unknown exception while processing rule "
                        "EYYYY: \"'Bad template'\""
                    ),
                )
            ],
            matches,
        )

    def test_parent_child(self):
        rules = Rules()
        rules.extend([RuleParent(), RuleChild()])

        cfn = Template(
            "-", dict_node({}, Mark(0, 0), Mark(0, 0)), regions=["us-east-1"]
        )

        matches = list(rules.run("-", cfn, ConfigMixIn([])))
        self.assertListEqual(
            [
                Match(
                    linenumber=1,
                    columnnumber=1,
                    linenumberend=1,
                    columnnumberend=1,
                    filename="-",
                    rule=RuleParent(),
                    message="Parent Rule",
                ),
                Match(
                    linenumber=1,
                    columnnumber=1,
                    linenumberend=1,
                    columnnumberend=1,
                    filename="-",
                    rule=RuleChild(),
                    message="Child Rule",
                ),
            ],
            matches,
            matches,
        )

    def test_parent_child_ignore(self):
        rules = Rules()
        rules.extend([RuleParent(), RuleChild()])

        cfn = Template(
            "-", dict_node({}, Mark(0, 0), Mark(0, 0)), regions=["us-east-1"]
        )

        matches = list(
            rules.run(
                "-", cfn, ConfigMixIn(ignore_checks=["E"], mandatory_checks=["ECCCC"])
            )
        )
        self.assertListEqual(
            [
                Match(
                    linenumber=1,
                    columnnumber=1,
                    linenumberend=1,
                    columnnumberend=1,
                    filename="-",
                    rule=RuleChild(),
                    message="Child Rule",
                )
            ],
            matches,
            matches,
        )

    def test_rule_deletion(self):
        rules = Rules()
        rules.extend([Rule()])

        with self.assertRaises(RuntimeError):
            del rules["EXXXX"]

    def test_failure_on_duplicate_rule(self):
        rules = Rules()

        with self.assertRaises(DuplicateRuleError):
            rules.extend([Rule(), Rule()])

    def test_custom_rules(self):
        rules = Rules.create_from_custom_rules_file(
            "test/fixtures/custom_rules/good/custom_rule_perfect.txt"
        )

        self.assertEqual(len(rules), 15)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context._mappings import Mappings
from cfnlint.context.context import Context
from cfnlint.jsonschema import ValidationError
from cfnlint.jsonschema.validators import CfnTemplateValidator


def _resolve(name, instance, expected_results, **kwargs):
    validator = CfnTemplateValidator().evolve(**kwargs)

    resolutions = list(validator.resolve_value(instance))

    for i, (instance, v, errors) in enumerate(resolutions):
        assert instance == expected_results[i][0]
        assert v.context.path.value_path == expected_results[i][1]
        assert errors == expected_results[i][2]


@pytest.mark.parametrize(
    "name,instance,response",
    [
        (
            "Valid Ref with a single instance",
            {"Ref": {"Ref": "MyResource"}},
            [],
        ),
    ],
)
def test_resolvers_ref(name, instance, response):
    _resolve(name, instance, response)


@pytest.mark.parametrize(
    "name,instance,response",
    [
        (
            "Invalid Ref with an array",
            {"Ref": []},
            [],
        ),
        (
            "Invalid Ref with an invalid object",
            {"Ref": {"foo": "bar"}},
            [],
        ),
        (
            "Invalid Join with an invalid type",
            {"Fn::Join": {"foo": "bar"}},
            [],
        ),
        (
            "Invalid Join with an invalid array length",
            {"Fn::Join": ["a", "b", "c"]},
            [],
        ),
        (
            "Invalid Join with an invalid type for first element",
            {"Fn::Join": [["a"], "b"]},
            [],
        ),
        (
            "Invalid Join with an invalid type for second element",
            {"Fn::Join": ["a", "b"]},
            [],
        ),
        (
            "Invalid Join with an invalid type for second element item",
            {"Fn::Join": ["/", [["a"], "b"]]},
            [],
        ),
        (
            "Invalid Select with an invalid type",
            {"Fn::Select": {"foo": "bar"}},
            [],
        ),
        (
            "Invalid Select with an invalid array length",
            {"Fn::Select": ["a", "b", "c"]},
            [],
        ),
        (
            "Invalid Select with an invalid type for second element",
            {"Fn::Select": ["a", "b"]},
            [],
        ),
        (
            "Invalid Select with an invalid type for second element",
            {"Fn::Select": [0, "b"]},
            [],
        ),
        (
            "Invalid Split with an invalid type",
            {"Fn::Split": {"foo": "bar"}},
            [],
        ),
        (
            "Invalid Split with an invalid array length",
            {"Fn::Split": ["a", "b", "c"]},
            [],
        ),
        (
            "Invalid Split with an invalid type for first element",
            {"Fn::Split": [0, "b"]},
            [],
        ),
        (
            "Invalid Split with an invalid type for second element",
            {"Fn::Split": ["a", ["b"]]},
            [],
        ),
        (
            "Invalid GetAZs with an invalid type ",
            {"Fn::GetAZs": {"a": "b"}},
            [],
        ),
        (
            "Invalid GetAZs with an invalid type ",
            {"Fn::GetAZs": ["a", "b"]},
            [],
        ),
        (
            "Invalid GetAZs with an invalid region",
            {"Fn::GetAZs": "foo"},
            [],
        ),
        (
            "Invalid FindInMap with an invalid type",
            {"Fn::FindInMap": {"foo": "bar"}},
            [],
        ),
        (
            "Invalid FindInMap with an invalid length",
            {"Fn::FindInMap": ["foo", "bar"]},
            [],
        ),
        (
            "Invalid FindInMap with an invalid type for first element",
            {"Fn::FindInMap": [["foo"], "bar", "value"]},
            [],
        ),
        (
            "Invalid FindInMap with an invalid type for second element",
            {"Fn::FindInMap": ["foo", ["bar"], "value"]},
            [],
        ),
        (
            "Invalid FindInMap with an invalid type for third element",
            {"Fn::FindInMap": ["foo", "bar", ["value"]]},
            [],
        ),
        (
            "Invalid FnSub with an invalid type",
            {"Fn::Sub": {"foo": "bar"}},
            [],
        ),
        (
            "Invalid FnSub with an array of wrong length",
            {"Fn::Sub": ["foo", "bar", "value"]},
            [],
        ),
        (
            "Invalid FnSub with the wrong type for element one",
            {"Fn::Sub": [["foo"], {"foo": "bar"}]},
            [],
        ),
        (
            "Invalid FnSub with the wrong type for element two",
            {"Fn::Sub": ["foo", ["bar"]]},
            [],
        ),
    ],
)
def test_invalid_functions(name, instance, response):
    context = Context(
        mappings=Mappings.create_from_dict({"foo": {"first": {"second": "bar"}}})
    )

    _resolve(name, instance, response, context=context)


@pytest.mark.parametrize(
    "name,instance,response",
    [
        (
            "Valid Join with an empty type",
            {"Fn::Join": [".", []]},
            [],
        ),
        (
            "Valid Join with a list of strings",
            {"Fn::Join": [".", ["a", "b", "c"]]},
            [("a.b.c", deque([]), None)],
        ),
        (
            "Valid GetAZs with empty string",
            {"Fn::GetAZs": ""},
            [
                (
                    [
                        "us-east-1a",
                        "us-east-1b",
                        "us-east-1c",
                        "us-east-1d",
                        "us-east-1e",
                        "us-east-1f",
                    ],
                    deque([]),
                    None,
                )
            ],
        ),
        (
            "Valid FindInMap with a default value",
            {"Fn::FindInMap": ["foo", "bar", "value", {"DefaultValue": "default"}]},
            [("default", deque([4, "DefaultValue"]), None)],
        ),
        (
            "Valid FindInMap with a default value",
            {"Fn::FindInMap": ["foo", "first", "second", {"DefaultValue": "default"}]},
            [
                ("default", deque([4, "DefaultValue"]), None),
                ("bar", deque([2]), None),
            ],
        ),
        (
            "Valid FindInMap with a bad mapping",
            {"Fn::FindInMap": ["bar", "first", "second"]},
            [
                (
                    None,
                    deque([]),
                    ValidationError(
                        (
                            "'bar' is not one of ['foo', "
                            "'transformFirstKey', 'transformSecondKey']"
                        ),
                        path=deque(["Fn::FindInMap", 0]),
                    ),
                )
            ],
        ),
        (
            "Valid FindInMap with a bad mapping and default",
            {"Fn::FindInMap": ["bar", "first", "second", {"DefaultValue": "default"}]},
            [("default", deque([4, "DefaultValue"]), None)],
        ),
        (
            "Valid FindInMap with a bad mapping and aws no value",
            {
                "Fn::FindInMap": [
                    "bar",
                    "first",
                    "second",
                    {"DefaultValue": {"Ref": "AWS::NoValue"}},
                ]
            },
            [],
        ),
        (
            "Valid FindInMap with a bad top key",
            {"Fn::FindInMap": ["foo", "second", "first"]},
            [
                (
                    None,
                    deque([]),
                    ValidationError(
                        ("'second' is not one of ['first'] for " "mapping 'foo'"),
                        path=deque(["Fn::FindInMap", 1]),
                    ),
                )
            ],
        ),
        (
            "Valid FindInMap with a bad top key and default",
            {"Fn::FindInMap": ["foo", "second", "first", {"DefaultValue": "default"}]},
            [("default", deque([4, "DefaultValue"]), None)],
        ),
        (
            "Valid FindInMap with a bad third key",
            {"Fn::FindInMap": ["foo", "first", "third"]},
            [
                (
                    None,
                    deque([]),
                    ValidationError(
                        (
                            "'third' is not one of ['second'] for "
                            "mapping 'foo' and key 'first'"
                        ),
                        path=deque(["Fn::FindInMap", 2]),
                    ),
                )
            ],
        ),
        (
            "Valid FindInMap with a bad second key and default",
            {"Fn::FindInMap": ["foo", "first", "third", {"DefaultValue": "default"}]},
            [("default", deque([4, "DefaultValue"]), None)],
        ),
        (
            "Valid FindInMap with a transform on first key",
            {"Fn::FindInMap": ["transformFirstKey", "first", "third"]},
            [],
        ),
        (
            "Valid FindInMap with a transform on second key",
            {"Fn::FindInMap": ["transformSecondKey", "first", "third"]},
            [],
        ),
        (
            "Valid Sub with a resolvable values",
            {"Fn::Sub": ["${a}-${b}", {"a": "foo", "b": "bar"}]},
            [("foo-bar", deque([]), None)],
        ),
        (
            "Valid Sub with empty parameters",
            {"Fn::Sub": ["foo", {}]},
            [("foo", deque([]), None)],
        ),
    ],
)
def test_valid_functions(name, instance, response):
    context = Context(
        mappings=Mappings.create_from_dict(
            {
                "foo": {"first": {"second": "bar"}},
                "transformFirstKey": {"Fn::Transform": {"second": "bar"}},
                "transformSecondKey": {"first": {"Fn::Transform": "bar"}},
            }
        )
    )
    _resolve(name, instance, response, context=context)


@pytest.mark.parametrize(
    "name,instance,response",
    [
        (
            "Invalid FindInMap with no mappings",
            {"Fn::FindInMap": [{"Ref": "MyParameter"}, "B", "C"]},
            [
                (
                    None,
                    deque([]),
                    ValidationError(
                        ("{'Ref': 'MyParameter'} is not one of []"),
                        path=deque(["Fn::FindInMap", 0]),
                    ),
                )
            ],
        ),
        (
            "Invalid FindInMap with no mappings and default value",
            {"Fn::FindInMap": ["A", "B", "C", {"DefaultValue": "default"}]},
            [("default", deque([4, "DefaultValue"]), None)],
        ),
    ],
)
def test_no_mapping(name, instance, response):
    _resolve(name, instance, response)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from typing import Any

from cfnlint.helpers import FUNCTIONS
from cfnlint.jsonschema import Validator
from cfnlint.rules.jsonschema import CfnLintKeyword


class ImageId(CfnLintKeyword):
    id = "W2506"
    shortdesc = "Check if ImageId Parameters have the correct type"
    description = (
        "See if there are any refs for ImageId to a parameter "
        + "of inappropriate type. Appropriate Types are "
        + "[AWS::EC2::Image::Id, AWS::SSM::Parameter::Value<AWS::EC2::Image::Id>]"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/best-practices.html#parmtypes"
    tags = ["parameters", "ec2", "imageid"]

    def __init__(self):
        super().__init__(
            keywords=[
                "Resources/AWS::AutoScaling::LaunchConfiguration/Properties/ImageId",
                "Resources/AWS::Batch::ComputeEnvironment/Properties/ComputeResources/ImageId",
                "Resources/AWS::Cloud9::EnvironmentEC2/Properties/ImageId",
                "Resources/AWS::EC2::Instance/Properties/ImageId",
                "Resources/AWS::EC2::LaunchTemplate/Properties/LaunchTemplateData/ImageId",
                "Resources/AWS::EC2::SpotFleet/Properties/SpotFleetRequestConfigData/LaunchSpecifications/*/ImageId",
                "Resources/AWS::ImageBuilder::Image/Properties/ImageId",
            ]
        )
        self.parent_rules = ["E1020"]

    def validate(self, validator: Validator, _, instance: Any, schema: Any):
        if any(fn in validator.context.path.path for fn in FUNCTIONS):
            return

        value = instance.get("Ref")
        if value not in validator.context.parameters:
            return

        parameter_type = validator.context.parameters[value].type
        for err in validator.descend(
            instance=parameter_type,
            schema={
                "enum": [
                    "AWS::EC2::Image::Id",
                    "AWS::SSM::Parameter::Value<AWS::EC2::Image::Id>",
                ]
            },
        ):
            err.rule = self
            err.path_override = deque(["Parameters", value, "Type"])
            yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class DependsOn(CfnLintKeyword):
    """Check Base Resource Configuration"""

    id = "E3005"
    shortdesc = "Check DependsOn values for Resources"
    description = "Check that the DependsOn values are valid"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-attribute-dependson.html"
    tags = ["resources", "dependson"]

    def __init__(self):
        super().__init__(keywords=["Resources/*/DependsOn", "Resources/*/DependsOn/*"])

    def validate(self, validator: Validator, s: Any, instance: Any, schema: Any):
        if not validator.is_type(instance, "string"):
            return

        resources = list(validator.context.resources.keys())

        if len(validator.context.path.path) > 1:
            if (
                isinstance(validator.context.path.path[1], str)
                and validator.context.path.path[1] in resources
            ):
                resources.remove(validator.context.path.path[1])

        if instance not in resources:
            yield ValidationError(
                f"{instance!r} is not one of {resources!r}",
            )
            return

        if not validator.cfn:
            return

        for scenario in validator.cfn.is_resource_available(
            list(validator.context.path.path), instance
        ):
            if scenario:
                scenario_text = " and ".join(
                    [
                        f"when condition {k!r} is {scenario[k]!r}"
                        for k in sorted(scenario)
                    ]
                )

                yield ValidationError(f"{instance!r} will not exist {scenario_text}")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import json
from collections import deque
from typing import Any

import regex as re

from cfnlint.helpers import REGEX_DYN_REF
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.functions._BaseFn import BaseFn


class Split(BaseFn):
    """Check if Split values are correct"""

    id = "E1018"
    shortdesc = "Split validation of parameters"
    description = "Making sure the split function is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-split.html"
    tags = ["functions", "split"]

    def __init__(self) -> None:
        super().__init__("Fn::Split", ("array",), resolved_rule="W1033")

    def schema(self, validator: Validator, instance: Any) -> dict[str, Any]:
        return {
            "type": "array",
            "maxItems": 2,
            "minItems": 2,
            "fn_items": [
                {
                    "schema": {
                        "type": ["string"],
                    }
                },
                {
                    "functions": [
                        "Fn::Base64",
                        "Fn::FindInMap",
                        "Fn::GetAtt",
                        "Fn::GetAZs",
                        "Fn::If",
                        "Fn::ImportValue",
                        "Fn::Join",
                        "Fn::Select",
                        "Fn::Sub",
                        "Ref",
                    ],
                    "schema": {
                        "type": ["string"],
                    },
                },
            ],
        }

    def fn_split(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ) -> ValidationResult:
        errs = list(super().validate(validator, s, instance, schema))
        if errs:
            yield from iter(errs)
            return

        key, value = self.key_value(instance)
        if re.fullmatch(REGEX_DYN_REF, json.dumps(value[1])):
            yield ValidationError(
                f"{key!r} does not support dynamic references",
                validator=self.fn.py,
                path=deque([key, 1]),
            )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from test.unit.rules import BaseRuleTestCase

from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.resources.Configuration import Configuration


class TestResourceConfiguration(BaseRuleTestCase):
    """Test AWS Types"""

    def setUp(self):
        """Setup"""
        super(TestResourceConfiguration, self).setUp()
        self.rule = Configuration()

    def test_configurations(self):
        validator = CfnTemplateValidator({})
        errors = list(
            self.rule.validate(validator, "cfnResources", {"foo": {"Type": "bar"}}, {})
        )
        self.assertListEqual(
            errors,
            [],
            errors,
        )

        errors = list(self.rule.validate(validator, "cfnResources", {"foo": []}, {}))

        self.assertListEqual(
            errors,
            [
                ValidationError(
                    "[] is not of type 'object'",
                    rule=Configuration(),
                    path=deque(["foo"]),
                    schema_path=deque(["patternProperties", "^[a-zA-Z0-9]+$", "type"]),
                    validator="type",
                    validator_value="object",
                    instance=[],
                ),
            ],
            errors,
        )

        errors = list(
            self.rule.validate(validator, "cfnResources", {"foo": {"Type": []}}, {})
        )

        self.assertListEqual(
            errors,
            [
                ValidationError(
                    "[] is not of type 'string'",
                    rule=Configuration(),
                    path=deque(["foo", "Type"]),
                    schema_path=deque(
                        [
                            "patternProperties",
                            "^[a-zA-Z0-9]+$",
                            "properties",
                            "Type",
                            "type",
                        ]
                    ),
                    validator="type",
                    validator_value="string",
                    instance=[],
                ),
            ],
            errors,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import create_context_for_template
from cfnlint.context.context import Transforms
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules import CfnLintKeyword
from cfnlint.rules.functions.GetAtt import GetAtt
from cfnlint.template import Template


@pytest.fixture(scope="module")
def rule():
    rule = GetAtt()
    yield rule


@pytest.fixture(scope="module")
def cfn():
    return Template(
        "",
        {
            "Resources": {
                "MyBucket": {"Type": "AWS::S3::Bucket"},
                "MyCodePipeline": {"Type": "AWS::CodePipeline::Pipeline"},
            },
            "Parameters": {
                "MyResourceParameter": {"Type": "String", "Default": "MyBucket"},
                "MyAttributeParameter": {"Type": "String", "AllowedValues": ["Arn"]},
            },
        },
        regions=["us-east-1"],
    )


@pytest.fixture(scope="module")
def context(cfn):
    return create_context_for_template(cfn)


class _Pass(CfnLintKeyword):
    id = "AAAAA"

    def __init__(self) -> None:
        super().__init__(keywords=["*"])

    def validate(self, validator, s, instance, schema):
        return
        yield


class _Fail(CfnLintKeyword):
    id = "BBBBB"

    def __init__(self) -> None:
        super().__init__(keywords=["*"])

    def validate(self, validator, s, instance, schema):
        yield ValidationError("Fail")


@pytest.mark.parametrize(
    "name,instance,schema,context_evolve,child_rules,expected",
    [
        (
            "Valid GetAtt with a good attribute",
            {"Fn::GetAtt": ["MyBucket", "Arn"]},
            {"type": "string"},
            {},
            {},
            [],
        ),
        (
            "Invalid GetAtt with bad attribute",
            {"Fn::GetAtt": ["MyBucket", "foo"]},
            {"type": "string"},
            {},
            {},
            [
                ValidationError(
                    (
                        "'foo' is not one of ['Arn', 'DomainName', "
                        "'DualStackDomainName', 'RegionalDomainName', 'WebsiteURL']"
                    ),
                    path=deque(["Fn::GetAtt", 1]),
                    schema_path=deque([]),
                    validator="fn_getatt",
                ),
            ],
        ),
        (
            "Invalid GetAtt with bad resource name",
            {"Fn::GetAtt": ["Foo", "bar"]},
            {"type": "string"},
            {},
            {},
            [
                ValidationError(
                    "'Foo' is not one of ['MyBucket', 'MyCodePipeline']",
                    path=deque(["Fn::GetAtt", 0]),
                    schema_path=deque(["enum"]),
                    validator="fn_getatt",
                ),
            ],
        ),
        (
            "Invalid GetAtt with a bad type",
            {"Fn::GetAtt": {"foo": "bar"}},
            {"type": "string"},
            {},
            {},
            [
                ValidationError(
                    "{'foo': 'bar'} is not of type 'string', 'array'",
                    path=deque(["Fn::GetAtt"]),
                    schema_path=deque(["type"]),
                    validator="fn_getatt",
                ),
            ],
        ),
        (
            "Invalid GetAtt with a bad response type",
            {"Fn::GetAtt": "MyBucket.Arn"},
            {"type": "array"},
            {},
            {},
            [
                ValidationError(
                    "{'Fn::GetAtt': 'MyBucket.Arn'} is not of type 'array'",
                    path=deque(["Fn::GetAtt"]),
                    schema_path=deque(["type"]),
                    validator="fn_getatt",
                ),
            ],
        ),
        (
            "Invalid GetAtt with a bad response type and multiple types",
            {"Fn::GetAtt": "MyBucket.Arn"},
            {"type": ["array", "object"]},
            {},
            {},
            [
                ValidationError(
                    "{'Fn::GetAtt': 'MyBucket.Arn'} is not of type 'array', 'object'",
                    path=deque(["Fn::GetAtt"]),
                    schema_path=deque(["type"]),
                    validator="fn_getatt",
                ),
            ],
        ),
        (
            "Valid GetAtt with integer to string",
            {"Fn::GetAtt": "MyCodePipeline.Version"},
            {"type": ["integer"]},
            {
                "strict_types": False,
            },
            {},
            [],
        ),
        (
            "Invalid GetAtt with integer to string",
            {"Fn::GetAtt": "MyCodePipeline.Version"},
            {"type": ["integer"]},
            {
                "strict_types": True,
            },
            {},
            [
                ValidationError(
                    (
                        "{'Fn::GetAtt': 'MyCodePipeline.Version'} "
                        "is not of type 'integer'"
                    ),
                    path=deque(["Fn::GetAtt"]),
                    schema_path=deque(["type"]),
                    validator="fn_getatt",
                )
            ],
        ),
        (
            "Valid GetAtt with one good response type",
            {"Fn::GetAtt": "MyBucket.Arn"},
            {"type": ["array", "string"]},
            {},
            {},
            [],
        ),
        (
            "Valid Ref in GetAtt for resource",
            {"Fn::GetAtt": [{"Ref": "MyResourceParameter"}, "Arn"]},
            {"type": "string"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            {},
            [],
        ),
        (
            "Valid Ref in GetAtt for attribute",
            {"Fn::GetAtt": ["MyBucket", {"Ref": "MyAttributeParameter"}]},
            {"type": "string"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            {},
            [],
        ),
        (
            "Invalid Ref in GetAtt for attribute",
            {"Fn::GetAtt": ["MyBucket", {"Ref": "MyResourceParameter"}]},
            {"type": "string"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            {},
            [
                ValidationError(
                    (
                        "'MyBucket' is not one of ['Arn', 'DomainName', "
                        "'DualStackDomainName', 'RegionalDomainName', "
                        "'WebsiteURL'] when "
                        "{'Ref': 'MyResourceParameter'} is resolved"
                    ),
                    path=deque(["Fn::GetAtt", 1]),
                    validator="fn_getatt",
                )
            ],
        ),
        (
            "Invalid Ref in GetAtt for attribute",
            {"Fn::GetAtt": [{"Ref": "MyAttributeParameter"}, "Arn"]},
            {"type": "string"},
            {"transforms": Transforms(["AWS::LanguageExtensions"])},
            {},
            [
                ValidationError(
                    (
                        "'Arn' is not one of ['MyBucket', 'MyCodePipeline'] when "
                        "{'Ref': 'MyAttributeParameter'} is resolved"
                    ),
                    path=deque(["Fn::GetAtt", 0]),
                    schema_path=deque(["enum"]),
                    validator="fn_getatt",
                )
            ],
        ),
        (
            "Valid GetAtt with child rules",
            {"Fn::GetAtt": ["MyBucket", "Arn"]},
            {"type": "string"},
            {},
            {
                "AAAAA": _Pass(),
                "BBBBB": _Fail(),
                "CCCCC": None,
            },
            [ValidationError("Fail")],
        ),
    ],
)
def test_validate(
    name, instance, schema, context_evolve, child_rules, expected, rule, context, cfn
):
    context = context.evolve(**context_evolve)
    rule.child_rules = child_rules
    validator = CfnTemplateValidator({}, context=context, cfn=cfn)
    errs = list(rule.fn_getatt(validator, schema, instance, {}))

    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import cfnlint.data.schemas.other.template
from cfnlint.rules import CfnLintJsonSchema, SchemaDetails


class Description(CfnLintJsonSchema):
    """Check Template Description is only a String"""

    id = "E1004"
    shortdesc = "Template description can only be a string"
    description = "Template description can only be a string"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/template-description-structure.html"
    tags = ["description"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Description"],
            schema_details=SchemaDetails(
                cfnlint.data.schemas.other.template,
                "description.json",
            ),
            all_matches=True,
        )
        self.rule_set = {
            "maxLength": "E1003",
        }
        self.child_rules = dict.fromkeys(list(self.rule_set.values()))



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.formats import FormatKeyword
from cfnlint.rules.formats.Format import Format


@pytest.fixture(scope="module")
def rule():
    rule = Format()
    yield rule


class _Pass(FormatKeyword):
    id = "AAAAA"

    def __init__(self):
        super().__init__(format="test")

    def format(self, validator, instance):
        return True


class _Fail(FormatKeyword):
    id = "BBBBB"

    def __init__(self):
        super().__init__(format="test")

    def format(self, validator, instance):
        return False


class _NoRuleId(FormatKeyword):
    def __init__(self):
        super().__init__(format="test")

    def format(self, validator, instance):
        return False


@pytest.mark.parametrize(
    "name,format,instance,child_rules,expected",
    [
        (
            "Standard format",
            "ipv4",
            "10.10.10.10",
            [],
            [],
        ),
        (
            "Bad standard format",
            "ipv4",
            "10.10.10",
            [],
            [
                ValidationError(
                    "'10.10.10' is not a 'ipv4'",
                )
            ],
        ),
        (
            "Pass format type",
            "test",
            "10.10.10.10",
            {
                "AAAAA": _Pass(),
            },
            [],
        ),
        (
            "Child rule empty",
            "test",
            "10.10.10.10",
            {
                "BBBBB": None,
            },
            [],
        ),
        (
            "No Rule ID",
            "test",
            "10.10.10.10",
            {
                "AAAAA": _NoRuleId(),
            },
            [],
        ),
        (
            "Fail with new format type",
            "test",
            "10.10.10.10",
            {
                "BBBBB": _Fail(),
            },
            [
                ValidationError(
                    "'10.10.10.10' is not a 'test'",
                    rule=_Fail(),
                )
            ],
        ),
        (
            "Fail with multiple rules",
            "test",
            "10.10.10.10",
            {
                "BBBBB": _Fail(),
                "AAAAA": _Pass(),
            },
            [
                ValidationError(
                    "'10.10.10.10' is not a 'test'",
                    rule=_Fail(),
                )
            ],
        ),
    ],
)
def test_validate(name, format, instance, child_rules, expected, rule, validator):
    rule.child_rules = child_rules
    result = list(rule.format(validator, format, instance, {}))
    assert result == expected, f"Test {name!r} got {result!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_lambda_function
from cfnlint.jsonschema import ValidationResult
from cfnlint.jsonschema.protocols import Validator
from cfnlint.rules.jsonschema import CfnLintJsonSchema, SchemaDetails


class FunctionEnvironmentKeys(CfnLintJsonSchema):

    id = "E3663"
    shortdesc = "Validate Lambda environment variable names aren't reserved"
    description = (
        "Lambda reserves a set of environment variable names for its use. "
        "This rule validates that the provided environment variable names "
        "don't use the reserved variable names"
    )
    source_url = (
        "https://docs.aws.amazon.com/lambda/latest/dg/configuration-envvars.html"
    )
    tags = ["resources", "lambda", "runtime"]

    def __init__(self):
        """Init"""
        super().__init__(
            keywords=[
                "Resources/AWS::Lambda::Function/Properties/Environment/Variables"
            ],
            schema_details=SchemaDetails(
                cfnlint.data.schemas.extensions.aws_lambda_function,
                "environment_variable_keys.json",
            ),
            all_matches=True,
        )

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        for err in super().validate(validator, keywords, instance, schema):
            err.message = (
                f"{err.instance!r} is a reserved variable name, one of "
                f"{self.schema.get('propertyNames').get('not').get('enum')!r}"
            )
            yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import Path
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.properties.PropertiesTemplated import PropertiesTemplated
from cfnlint.template import Template


@pytest.fixture(scope="module")
def rule():
    rule = PropertiesTemplated()
    yield rule


@pytest.fixture
def path():
    return Path(
        path=deque(["Resources/AWS::CloudFormation::Template/Properties/TemplateURL"])
    )


@pytest.mark.parametrize(
    "name,instance,transforms,expected",
    [
        (
            "A s3 string should be fine",
            "s3://my-bucket/key/value.yaml",
            [],
            [],
        ),
        (
            "An object isn't a string so we skip",
            {"foo": "bar"},
            [],
            [],
        ),
        (
            "A transform will result in non failure",
            "./folder",
            "AWS::Serverless-2016-10-31",
            [],
        ),
        (
            "String with no transform",
            "./folder",
            [],
            [
                ValidationError(
                    ("This code may only work with 'package' cli command"),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, transforms, expected, rule, validator):
    validator = validator.evolve(cfn=Template("", {"Transform": transforms}))

    errs = list(rule.validate(validator, {}, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.helpers import TEMPLATED_PROPERTY_CFN_PATHS, ensure_list
from cfnlint.jsonschema._keywords_cfn import cfn_type
from cfnlint.rules import CloudFormationLintRule


class Type(CloudFormationLintRule):
    """Check if Resource PrimitiveTypes are correct"""

    id = "E3012"
    shortdesc = "Check resource properties values"
    description = (
        "Checks resource property values with Primitive Types for values that match"
        " those types."
    )
    source_url = "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/cfn-schema-specification.md#type"
    tags = ["resources"]

    strict_exceptions = {
        "AWS::CloudFormation::Stack": ["Parameters"],
        "AWS::Lambda::Function.Environment": ["Variables"],
    }

    def __init__(self):
        """Init"""
        super().__init__()
        self.config_definition = {
            "strict": {"default": False, "type": "boolean"},
        }  # Strict mode is set to false by default
        self.configure()

    # pylint: disable=unused-argument
    def type(self, validator, types, instance, schema):
        if validator.context.path.cfn_path:
            if (
                validator.is_type(instance, "string")
                and not validator.context.strict_types
            ):
                if (
                    "/".join(validator.context.path.cfn_path)
                    in TEMPLATED_PROPERTY_CFN_PATHS
                ):
                    return

        if (
            self.config.get("strict") or validator.context.strict_types
        ) and not validator.context.is_resolved_value:
            validator = validator.evolve(
                context=validator.context.evolve(strict_types=True)
            )
        else:
            validator = validator.evolve(
                context=validator.context.evolve(strict_types=False)
            )

        types = ensure_list(types)
        for err in cfn_type(
            validator,
            types,
            instance,
            schema,
        ):
            for t in ["object", "array", "boolean", "integer", "number", "string"]:
                if validator.is_type(instance, t):
                    err.extra_args = {
                        "actual_type": t,
                        "expected_type": types[0],
                    }
                    yield err
                    break



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

import cfnlint.data.schemas.extensions.aws_s3_bucket
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class AccessControlOwnership(CfnLintJsonSchema):
    """Check if the S3 Bucket Access Controls are set with Ownership rules"""

    id = "E3045"
    shortdesc = "Validate AccessControl are set with OwnershipControls"
    description = (
        "When using AccessControl other than private you must also "
        "configure OwnershipControls. The default is bucket owner "
        "enforced which disables ACLs."
    )
    source_url = "https://docs.aws.amazon.com/AmazonS3/latest/userguide/about-object-ownership.html"
    tags = ["resources", "s3"]

    def __init__(self):
        super().__init__(
            keywords=["Resources/AWS::S3::Bucket/Properties"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_s3_bucket,
                filename="ownershipcontrols.json",
            ),
        )

    def message(self, instance: Any, err: ValidationError) -> str:
        return (
            "A bucket with 'AccessControl' set should also have "
            "at least one 'OwnershipControl' configured"
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule


class SubResolved(CloudFormationLintRule):
    id = "W1031"
    shortdesc = "Validate the values that come from a Fn::Sub function"
    description = "Resolve the Fn::Sub and then validate the values against the schema"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-sub.html"
    tags = ["functions", "sub"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import Path, create_context_for_template
from cfnlint.helpers import FUNCTIONS
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.resources.properties.AvailabilityZone import (
    AvailabilityZone,  # pylint: disable=E0401
)
from cfnlint.template import Template


@pytest.fixture(scope="module")
def rule():
    rule = AvailabilityZone()
    yield rule


@pytest.fixture
def template():
    return {
        "Resources": {
            "MySqs": {
                "Type": "AWS::SQS::Queue",
            }
        }
    }


@pytest.fixture(scope="module")
def validator_cdk():
    cfn = Template(
        "",
        {
            "Resources": {
                "MySqs": {
                    "Type": "AWS::SQS::Queue",
                },
                "CDK": {
                    "Type": "AWS::CDK::Metadata",
                },
            }
        },
    )
    context = (
        create_context_for_template(cfn)
        .evolve(
            functions=FUNCTIONS,
            path="Resources",
        )
        .evolve(path="MySqs")
        .evolve(path="Properties")
    )
    yield CfnTemplateValidator(schema={}, context=context, cfn=cfn)


@pytest.mark.parametrize(
    "name,instance,path,expected",
    [
        (
            "Valid ref",
            {"Ref": "AZ"},
            {
                "path": deque(["Resources", "MySqs", "Properties"]),
            },
            [],
        ),
        (
            "Valid list ref",
            [{"Ref": "AZ"}],
            {
                "path": deque(["Resources", "MySqs", "Properties"]),
            },
            [],
        ),
        (
            "Valid inside Ref",
            "us-east-1a",
            {"path": deque(["Ref"])},
            [],
        ),
        (
            "Valid GetAZs",
            ["us-east-1a", "us-east-1b"],
            {"path": deque(["Fn::GetAZs"])},
            [],
        ),
        (
            "Invalid type",
            True,
            {
                "path": deque(["Resources", "MySqs", "Properties"]),
            },
            [],
        ),
        (
            "Valid hardcoded all string",
            "all",
            {
                "path": deque(["Resources", "MySqs", "Properties"]),
            },
            [],
        ),
        (
            "Invalid hardcoded string",
            "us-east-1a",
            {
                "path": deque(["Resources", "MySqs", "Properties"]),
            },
            [
                ValidationError(
                    ("Avoid hardcoding availability zones 'us-east-1a'"),
                    rule=AvailabilityZone(),
                )
            ],
        ),
        (
            "Invalid hardcoded array",
            ["us-east-1a", "us-east-1b"],
            {
                "path": deque(["Resources", "MySqs", "Properties"]),
            },
            [
                ValidationError(
                    ("Avoid hardcoding availability zones 'us-east-1a'"),
                    rule=AvailabilityZone(),
                ),
                ValidationError(
                    ("Avoid hardcoding availability zones 'us-east-1b'"),
                    rule=AvailabilityZone(),
                ),
            ],
        ),
    ],
    indirect=["path"],
)
def test_validate(name, instance, path, expected, rule, validator):
    errors = list(rule.validate(validator, False, instance, {}))
    # we use error counts in this one as the instance types are
    # always changing so we aren't going to hold ourselves up by that
    assert errors == expected, f"Test {name!r} got {errors!r}"


@pytest.mark.parametrize(
    "name,instance,path,expected",
    [
        (
            "Valid hardcoded string because CDK",
            "us-east-1a",
            deque(["Resources", "MySqs", "Properties"]),
            [],
        ),
    ],
)
def test_validate_for_cdk(name, instance, path, expected, rule, validator_cdk):
    validator = validator_cdk.evolve(
        context=validator_cdk.context.evolve(
            path=Path(path=path),
        )
    )
    errors = list(rule.validate(validator, False, instance, {}))
    # we use error counts in this one as the instance types are
    # always changing so we aren't going to hold ourselves up by that
    assert errors == expected, f"Test {name!r} got {errors!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.data import AdditionalSpecs
from cfnlint.helpers import ensure_list, load_resource
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class Permissions(CfnLintKeyword):
    """Check IAM Permission configuration"""

    id = "W3037"
    shortdesc = "Check IAM Permission configuration"
    description = "Check for valid IAM Permissions"
    source_url = "https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements_action.html"
    tags = ["properties", "iam", "permissions"]
    experimental = True

    def __init__(self):
        """Init"""
        super().__init__(
            ["AWS::IAM::Policy/Properties/PolicyDocument/Statement/Action"]
        )
        self.service_map = self.load_service_map()

    def validate(
        self, validator: Validator, _, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        actions = ensure_list(instance)

        for action in actions:
            if action == "*":
                continue
            if ":" not in action:
                yield ValidationError(
                    (
                        f"{action!r} is not a valid action."
                        "Must be of the form service:action or '*'"
                    ),
                    rule=self,
                )
                return
            service, permission = action.split(":", 1)
            service = service.lower()
            permission = permission.lower()

            if service in self.service_map:
                enums = self.service_map[service]
                if permission == "*":
                    pass
                elif permission.endswith("*"):
                    wilcarded_permission = permission.split("*")[0]
                    if not any(wilcarded_permission in action for action in enums):
                        yield ValidationError(
                            f"{permission!r} is not one of {enums!r}",
                            rule=self,
                        )

                elif permission.startswith("*"):
                    wilcarded_permission = permission.split("*")[1]
                    if not any(
                        wilcarded_permission in action
                        for action in self.service_map[service]
                    ):
                        yield ValidationError(
                            f"{permission!r} is not one of {enums!r}",
                            rule=self,
                        )
                elif permission not in self.service_map[service]:
                    yield ValidationError(
                        f"{permission!r} is not one of {enums!r}",
                        rule=self,
                    )
            else:
                yield ValidationError(
                    f"{service!r} is not one of {list(self.service_map.keys())!r}",
                    rule=self,
                )

    def load_service_map(self):
        """
        Convert policies.json into a simpler version for more efficient key lookup.
        """
        service_map = load_resource(AdditionalSpecs, "Policies.json")["serviceMap"]

        policy_service_map = {}

        for _, properties in service_map.items():
            # The services and actions are case insensitive
            service = properties["StringPrefix"].lower()
            actions = [x.lower() for x in properties["Actions"]]

            # Some services have the same name for different
            # generations; like elasticloadbalancing.
            if service in policy_service_map:
                policy_service_map[service] += actions
            else:
                policy_service_map[service] = actions

        return policy_service_map



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any, Iterator

from cfnlint.context.conditions import Unsatisfiable
from cfnlint.helpers import is_function
from cfnlint.jsonschema import Validator


def _get_relationship_fn_if(
    validator: Validator, key: Any, value: Any, path: deque[str | int]
) -> Iterator[tuple[Any, Validator]]:
    if not isinstance(value, list) or len(value) != 3:
        return
    condition = value[0]

    for i in [1, 2]:
        try:
            if_validator = validator.evolve(
                context=validator.context.evolve(
                    conditions=validator.context.conditions.evolve(
                        status={
                            condition: True if i == 1 else False,
                        },
                    ),
                    path=validator.context.path.descend(path=key).descend(path=i),
                )
            )
            for r, v in get_value_from_path(
                if_validator,
                value[i],
                path.copy(),
            ):
                yield r, v
        except Unsatisfiable:
            pass


def _get_value_from_path_list(
    validator: Validator, instance: Any, path: deque[str | int]
) -> Iterator[tuple[Any, Validator]]:
    for i, v in enumerate(instance):
        for r, v in get_value_from_path(
            validator.evolve(
                context=validator.context.evolve(
                    path=validator.context.path.descend(path=i)
                ),
            ),
            v,
            path.copy(),
        ):
            yield r, v


def get_value_from_path(
    validator: Validator, instance: Any, path: deque[str | int]
) -> Iterator[tuple[Any, Validator]]:
    """
    Retrieve a value from a nested dictionary or list using a path.

    Args:
        validator (Validator): The validator instance
        data (Any): The dictionary or list to search.
        path (deque[str | int]): The path to the value.

    Returns:
        The value at the specified path, or None if the key doesn't exist.

    Examples:
        >>> data = {'a': {'b': {'c': 3}}}
        >>> get_value_from_path(data, ['a', 'b', 'c'])
        3
    """

    fn_k, fn_v = is_function(instance)
    if fn_k is not None:
        if fn_k == "Fn::If":
            yield from _get_relationship_fn_if(validator, fn_k, fn_v, path)
        elif fn_k == "Ref" and fn_v == "AWS::NoValue":
            yield None, validator.evolve(
                context=validator.context.evolve(
                    path=validator.context.path.descend(path=fn_k)
                )
            )
        elif not path:
            yield instance, validator
        return

    if not path:
        yield instance, validator
        return

    key = path.popleft()
    if isinstance(instance, list) and key == "*":
        yield from _get_value_from_path_list(validator, instance, path)
        return

    if not isinstance(instance, dict):
        yield None, validator
        return

    for r, v in get_value_from_path(
        validator.evolve(
            context=validator.context.evolve(
                path=validator.context.path.descend(path=key)
            )
        ),
        instance.get(key),
        path.copy(),
    ):
        yield r, v

    return



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule


class DependentRequired(CloudFormationLintRule):
    """Check Required Resource Configuration"""

    id = "E3021"
    shortdesc = (
        "Validate that when a property is specified that "
        "other properties should be included"
    )
    description = (
        "When certain properties are specified it results "
        "in other properties to be required"
    )
    source_url = "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/cfn-schema-specification.md#dependentrequired"
    tags = ["resources"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from test.unit.rules import BaseRuleTestCase

from cfnlint._typing import Path
from cfnlint.decode.cfn_json import Mark
from cfnlint.decode.node import dict_node, str_node
from cfnlint.rules import CloudFormationLintRule, RuleMatch
from cfnlint.rules.jsonschema.JsonSchema import JsonSchema
from cfnlint.template import Template


class RuleWithFunction(CloudFormationLintRule):
    """Test Rule"""

    id = "E3XXX"
    shortdesc = "Test Rule"
    description = "Test Rule"
    source_url = ""
    tags: list[str] = []


class RuleWithOutFunction(CloudFormationLintRule):
    """Test Rule"""

    id = "E3YYY"
    shortdesc = "Test Rule"
    description = "Test Rule"
    source_url = ""
    tags: list[str] = []


class RuleRefFunction(CloudFormationLintRule):
    """Test Rule"""

    id = "E3YYY"
    shortdesc = "Test Rule"
    description = "Test Rule"
    source_url = ""
    tags: list[str] = []


class RuleFnIfFunction(CloudFormationLintRule):
    """Test Rule"""

    id = "E3AAAA"
    shortdesc = "Test Rule for Fn::IF function"
    description = "Test Rule"
    source_url = ""
    tags: list[str] = []


class TestJsonSchema(BaseRuleTestCase):
    """Test Json Size"""

    def setUp(self):
        """Setup"""
        super(TestJsonSchema, self).setUp()
        self.maxDiff = None
        self.rule = JsonSchema()
        self.template = build_dict({})
        self.cfn = Template("", self.template, ["us-east-1"])

    def build_result(self, message: str, path: Path) -> RuleMatch:
        if len(path) > 0:
            return RuleMatch(
                path[:],
                message,
                rule=self.rule,
                location=(
                    f"{path[-1]}-sm-line",
                    f"{path[-1]}-sm-column",
                    f"{path[-1]}-em-line",
                    f"{path[-1]}-em-column",
                ),
            )

        return RuleMatch(path[:], message, rule=self.rule, location=(1, 1, 1, 1))

    def validate(self, expected):
        matches = self.rule.match(self.cfn)

        self.assertListEqual(
            list(map(vars, expected)),
            list(map(vars, matches)),
            list(map(vars, matches)),
        )

    def test_required_properties(self):
        expected = [
            self.build_result(
                "'Resources' is a required property",
                [],
            ),
        ]

        self.cfn = Template("", build_dict({}), ["us-east-1"])
        delattr(expected[0], "location")
        self.validate(expected)

    def test_extra(self):
        expected = [
            self.build_result(
                "Additional properties are not allowed ('Foo' was unexpected)",
                ["Foo"],
            ),
        ]
        self.cfn = Template(
            "",
            build_dict(
                {
                    build_str("Resources"): build_dict({}),
                    build_str("Foo"): build_dict({}),
                },
            ),
            ["us-east-1"],
        )
        self.validate(expected)

    def test_config(self):
        self.rule.config["sections"] = "Foo"
        self.cfn = Template(
            "",
            build_dict(
                {
                    build_str("Resources"): build_dict({}),
                    build_str("Foo"): build_dict({}),
                },
            ),
            ["us-east-1"],
        )
        self.validate([])


def build_node(instance: str | dict, loc: str, type=str_node):
    return type(
        instance,
        Mark(f"{loc}-sm-line", f"{loc}-sm-column"),  # type: ignore
        Mark(f"{loc}-em-line", f"{loc}-em-column"),  # type: ignore
    )


def build_str(instance: str):
    return build_node(
        instance,
        instance,
        str_node,
    )


def build_dict(instance: dict):
    return build_node(
        instance,
        "object",
        dict_node,
    )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque

import pytest

from cfnlint.rules.helpers import get_resource_by_name


@pytest.fixture
def template():
    return {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Parameters": {
            "ImageId": {
                "Type": "String",
            }
        },
        "Conditions": {
            "IsUsEast1": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]},
            "IsNotUsEast1": {"Fn::Not": [{"Condition": "IsUsEast1"}]},
            "IsImageId": {"Fn::Not": {"Fn::Equals": [{"Ref": "ImageId"}, ""]}},
        },
        "Resources": {
            "Foo": {"Type": "AWS::S3::Bucket"},
            "Bar": {"Type": "AWS::S3::Bucket", "Condition": "IsUsEast1"},
            "FooBar": {"Type": "AWS::S3::Bucket", "Condition": "IsNotUsEast1"},
            "BadShape": [],
            "NoType": {},
        },
    }


@pytest.mark.parametrize(
    "name,resource_name,types,status,expected",
    [
        (
            "Standard get",
            "Foo",
            None,
            {},
            ({"Type": "AWS::S3::Bucket"}, {}, deque(["Resources", "Foo"])),
        ),
        (
            "Doesn't exist",
            "Foo2",
            None,
            {},
            (
                None,
                {},
                deque([]),
            ),
        ),
        (
            "The destination isn't a dict",
            "BadShape",
            None,
            {},
            (
                None,
                {},
                deque([]),
            ),
        ),
        (
            "The destination doesn't have type",
            "NoType",
            None,
            {},
            (
                None,
                {},
                deque([]),
            ),
        ),
        (
            "Get a valid resource with a filter",
            "Foo",
            ["AWS::S3::Bucket"],
            {},
            (
                {
                    "Type": "AWS::S3::Bucket",
                },
                {},
                deque(["Resources", "Foo"]),
            ),
        ),
        (
            "No result when type filters it out",
            "Foo",
            ["AWS::EC2::Instance"],
            {},
            (
                None,
                {},
                deque([]),
            ),
        ),
        (
            "Get a resource with a condition",
            "Bar",
            None,
            {},
            (
                {"Type": "AWS::S3::Bucket", "Condition": "IsUsEast1"},
                {"IsUsEast1": True},
                deque(["Resources", "Bar"]),
            ),
        ),
        (
            "No resource when conditions don't align 1",
            "Bar",
            None,
            {"IsUsEast1": False},
            (
                None,
                {"IsUsEast1": False},
                deque([]),
            ),
        ),
        (
            "No resource when conditions don't align 2",
            "FooBar",
            None,
            {"IsUsEast1": True},
            (
                None,
                {"IsUsEast1": True},
                deque([]),
            ),
        ),
        (
            "Unrelated conditions return full results",
            "Bar",
            None,
            {"IsImageId": False},
            (
                {"Type": "AWS::S3::Bucket", "Condition": "IsUsEast1"},
                {"IsUsEast1": True, "IsImageId": False},
                deque(["Resources", "Bar"]),
            ),
        ),
    ],
)
def test_get_resource_by_name(name, resource_name, types, status, expected, validator):
    validator = validator.evolve(
        context=validator.context.evolve(
            conditions=validator.context.conditions.evolve(
                status=status,
            ),
        ),
    )

    result = get_resource_by_name(validator, resource_name, types)

    assert result[0] == expected[0], f"Test {name!r} got {result[0]!r}"
    assert (
        result[1].context.conditions.status == expected[1]
    ), f"Test {name!r} got {result[1].context.conditions.status!r}"
    assert (
        result[1].context.path.path == expected[2]
    ), f"Test {name!r} got {result[1].context.path.path!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from test.unit.rules import BaseRuleTestCase

from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.conditions.Configuration import Configuration


class TestConfiguration(BaseRuleTestCase):
    def test_condition(self):
        rule = Configuration()
        validator = CfnTemplateValidator({})

        errors = list(rule.validate(validator, {}, {"foo": True}, []))
        self.assertListEqual(
            errors,
            [],
        )

        errors = list(rule.validate(validator, {}, {"bar": "bad"}, []))
        self.assertListEqual(
            errors,
            [
                ValidationError(
                    "'bad' is not of type 'boolean'",
                    path=deque(["bar"]),
                    rule=Configuration(),
                    schema_path=deque(
                        ["patternProperties", "^[a-zA-Z0-9&]{1,255}$", "type"]
                    ),
                    validator="type",
                    validator_value="boolean",
                    instance="bad",
                ),
            ],
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule


class ArrayLength(CloudFormationLintRule):
    """Check if List has a size within the limit"""

    id = "E3032"
    shortdesc = "Check if a array has between min and max number of values specified"
    description = (
        "Check array for the number of items in the list to validate they are between"
        " the minimum and maximum"
    )
    source_url = "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/cfn-schema-specification.md#arraylength"
    tags = ["resources", "property", "array", "length"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.testlib.testcase import BaseTestCase

import regex as re

import cfnlint.decode.cfn_yaml
from cfnlint.config import _DEFAULT_RULESDIR
from cfnlint.exceptions import DuplicateRuleError
from cfnlint.match import Match
from cfnlint.rules import CloudFormationLintRule, RulesCollection
from cfnlint.rules.errors import RuleError
from cfnlint.template.template import Template


class RuleFail(CloudFormationLintRule):
    """Def Rule"""

    id = "EYYYY"
    shortdesc = "Test Rule"
    description = "Test Rule"
    source_url = "https://github.com/aws-cloudformation/cfn-lint/"
    tags = ["resources"]

    def match(self, cfn):
        raise KeyError("Bad template")


class TestRulesCollection(BaseTestCase):
    """Test Template RulesCollection in cfnlint"""

    def setUp(self):
        """SetUp template object"""
        self.rules = RulesCollection()
        self.rules.include_rules = ["I", "W", "E"]
        rulesdirs = [_DEFAULT_RULESDIR]
        for rulesdir in rulesdirs:
            self.rules.create_from_directory(rulesdir)

    def test_rule_ids_unique(self):
        """Test Rule IDs are Unique"""
        existing_rules = []
        for rule in self.rules:
            self.assertFalse(rule.id in existing_rules)
            existing_rules.append(rule.id)

    def test_rule_ids_are_formatted_correctly(self):
        """Test Rule IDs are property formmated"""
        for rule in self.rules:
            self.assertIn(rule.id[0], ["W", "I", "E"])
            self.assertEqual(len(rule.id), 5)
            self.assertTrue(isinstance(int(rule.id[1:]), int))

    def test_success_run(self):
        """Test Run Logic"""
        filename = "test/fixtures/templates/good/generic.yaml"
        template = cfnlint.decode.cfn_yaml.load(filename)
        cfn = Template(filename, template, ["us-east-1"])

        matches = []
        matches.extend(self.rules.run(filename, cfn))
        assert matches == []

    def test_fail_run(self):
        """Test failure run"""
        filename = "test/fixtures/templates/bad/generic.yaml"
        template = cfnlint.decode.cfn_yaml.load(filename)
        cfn = Template(filename, template, ["us-east-1"])
        expected_err_count = 38
        matches = []
        matches.extend(self.rules.run(filename, cfn))
        assert (
            len(matches) == expected_err_count
        ), "Expected {} failures, got {}".format(expected_err_count, len(matches))

    def test_success_filtering_of_rules_default(self):
        """Test extend function"""

        class rule_e0000(CloudFormationLintRule):
            """Error Rule"""

            id = "E0000"

        class rule_w0000(CloudFormationLintRule):
            """Warning Rule"""

            id = "W0000"

        class rule_i0000(CloudFormationLintRule):
            """Info Rule"""

            id = "I0000"

        rules_to_add = [rule_e0000(), rule_w0000(), rule_i0000()]
        rules = RulesCollection(ignore_rules=None, include_rules=None)
        rules.extend(rules_to_add)
        self.assertEqual(len(rules), 2)
        for rule in rules:
            self.assertIn(rule.id, ["W0000", "E0000"])

    def test_success_filtering_of_rules_include_info(self):
        """Test extend function"""

        class rule_e0000(CloudFormationLintRule):
            """Error Rule"""

            id = "E0000"

        class rule_w0000(CloudFormationLintRule):
            """Warning Rule"""

            id = "W0000"

        class rule_i0000(CloudFormationLintRule):
            """Info Rule"""

            id = "I0000"

        rules_to_add = [rule_e0000(), rule_w0000(), rule_i0000()]
        rules = RulesCollection(ignore_rules=None, include_rules=["I"])
        rules.extend(rules_to_add)
        self.assertEqual(len(rules), 3)
        for rule in rules:
            self.assertIn(rule.id, ["I0000", "W0000", "E0000"])

    def test_success_filtering_of_rules_exclude(self):
        """Test extend function"""

        class rule_e0000(CloudFormationLintRule):
            """Error Rule"""

            id = "E0000"

        class rule_w0000(CloudFormationLintRule):
            """Warning Rule"""

            id = "W0000"

        class rule_i0000(CloudFormationLintRule):
            """Info Rule"""

            id = "I0000"

        rules_to_add = [rule_e0000(), rule_w0000(), rule_i0000()]
        rules = RulesCollection(ignore_rules=["E"])
        rules.extend(rules_to_add)
        self.assertEqual(len(rules), 1)
        for rule in rules:
            self.assertIn(rule.id, ["W0000"])

    def test_success_filtering_of_rules_exclude_long(self):
        """Test extend function"""

        class rule_e0000(CloudFormationLintRule):
            """Error Rule"""

            id = "E0000"

        class rule_e0010(CloudFormationLintRule):
            """Error Rule"""

            id = "E0010"

        class rule_e0002(CloudFormationLintRule):
            """Error Rule"""

            id = "E0002"

        rules_to_add = [rule_e0000(), rule_e0010(), rule_e0002()]
        rules = RulesCollection(ignore_rules=["E000"])
        rules.extend(rules_to_add)
        self.assertEqual(len(rules), 1)
        for rule in rules:
            self.assertIn(rule.id, ["E0010"])

    def test_success_filtering_of_rules_exclude_longer(self):
        """Test extend function"""

        class rule_e0000(CloudFormationLintRule):
            """Error Rule"""

            id = "E0000"

        class rule_e0010(CloudFormationLintRule):
            """Error Rule"""

            id = "E0010"

        class rule_e0002(CloudFormationLintRule):
            """Error Rule"""

            id = "E0002"

        rules_to_add = [rule_e0000(), rule_e0010(), rule_e0002()]
        rules = RulesCollection(ignore_rules=["E0002"])
        rules.extend(rules_to_add)
        self.assertEqual(len(rules), 2)
        for rule in rules:
            self.assertIn(rule.id, ["E0000", "E0010"])

    def test_success_filtering_of_rules_exclude_mandatory(self):
        """Test extend function"""

        class rule_e0000(CloudFormationLintRule):
            """Error Rule"""

            id = "E0000"

        class rule_w0000(CloudFormationLintRule):
            """Warning Rule"""

            id = "W0000"

        class rule_i0000(CloudFormationLintRule):
            """Info Rule"""

            id = "I0000"

        rules_to_add = [rule_e0000(), rule_w0000(), rule_i0000()]
        rules = RulesCollection(ignore_rules=["E"], mandatory_rules=["E"])
        rules.extend(rules_to_add)
        self.assertEqual(len(rules), 2)
        for rule in rules:
            self.assertIn(rule.id, ["E0000", "W0000"])

    def test_success_filtering_of_rules_exclude_mandatory_long(self):
        """Test extend function"""

        class rule_e0000(CloudFormationLintRule):
            """Error Rule"""

            id = "E0000"

        class rule_e0010(CloudFormationLintRule):
            """Error Rule"""

            id = "E0010"

        class rule_e0002(CloudFormationLintRule):
            """Error Rule"""

            id = "E0002"

        class rule_w0000(CloudFormationLintRule):
            """Warning Rule"""

            id = "W0000"

        rules_to_add = [rule_e0000(), rule_e0010(), rule_e0002(), rule_w0000()]
        rules = RulesCollection(ignore_rules=["E"], mandatory_rules=["E000"])
        rules.extend(rules_to_add)
        self.assertEqual(len(rules), 3)
        for rule in rules:
            self.assertIn(rule.id, ["E0000", "E0002", "W0000"])

    def test_duplicate_rules(self):
        """Test extend function"""

        class rule0_e0000(CloudFormationLintRule):
            """Error Rule"""

            id = "E0000"

        class rule1_e0000(CloudFormationLintRule):
            """Error Rule"""

            id = "E0000"

        rules_to_add = [rule0_e0000(), rule1_e0000()]
        rules = RulesCollection()
        self.assertRaises(DuplicateRuleError, rules.extend, rules_to_add)

    def test_repr(self):
        class rule0_e0000(CloudFormationLintRule):
            """Error Rule"""

            id = "E0000"
            shortdesc = "Rule A"
            description = "First rule"

        class rule1_e0001(CloudFormationLintRule):
            """Error Rule"""

            id = "E0001"
            shortdesc = "Rule B"
            description = "Second rule"

        rules = RulesCollection()
        rules.extend([rule0_e0000(), rule1_e0001()])

        retval = repr(rules)
        pattern = r"\AE0000: Rule A\nFirst rule\nE0001: Rule B\nSecond rule\Z"
        match = re.match(pattern, retval)
        assert match, f"{retval} does not match {pattern}"

    def test_custom_rules(self):
        rules = RulesCollection([], [])

        rules.create_from_custom_rules_file(
            "test/fixtures/custom_rules/good/custom_rule_perfect.txt"
        )

        self.assertEqual(len(rules), 15)

    def test_rule_failure(self):
        rules = RulesCollection(ignore_rules=["E"], mandatory_rules=["E000", "EYYYY"])
        rules.extend([RuleFail()])

        cfn = Template("-", {}, regions=["us-east-1"])

        matches = list(rules.run("-", cfn))
        self.assertListEqual(
            [
                Match(
                    linenumber=1,
                    linenumberend=1,
                    columnnumber=1,
                    columnnumberend=2,
                    filename="-",
                    rule=RuleError(),
                    message=(
                        "Unknown exception while processing rule "
                        "EYYYY: 'Bad template'"
                    ),
                )
            ],
            matches,
        )


class TestCreateFromModule(BaseTestCase):
    """Test loading a rules collection from a module"""

    def test_create_from_module(self):
        """Load rules from a module"""
        rules = RulesCollection()
        rules.create_from_module("cfnlint.rules.jsonschema.JsonSchema")
        self.assertIn("E1001", (r.id for r in rules))



"""
Copyright (c) 2013 Julian Berman

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

SPDX-License-Identifier: MIT
"""

# Code is taken from jsonschema package and adapted CloudFormation use
# https://github.com/python-jsonschema/jsonschema

import itertools
from collections.abc import Mapping, Sequence

import regex as re


class Unset:
    """
    An as-of-yet unset attribute or unprovided default parameter.
    """

    def __repr__(self):
        return "<unset>"


def find_additional_properties(validator, instance, schema):
    """
    Return the set of additional properties for the given ``instance``.

    Weeds out properties that should have been validated by ``properties`` and
    / or ``patternProperties``.

    Assumes ``instance`` is dict-like already.
    """

    properties = schema.get("properties", {})
    patterns = "|".join(schema.get("patternProperties", {}))
    for property in instance:
        if property not in properties:
            if validator.is_type(property, "string"):
                if patterns and re.search(patterns, property):
                    continue
            yield property


def custom_msg(validator, schema):
    """
    Create an error message for custom validation.
    """
    if not isinstance(schema, dict):
        return None
    messages = schema.get("message")
    if isinstance(messages, dict):
        return messages.get(validator)

    return None


def _mapping_equal(one, two):
    """
    Check if two mappings are equal using the semantics of `equal`.
    """
    if len(one) != len(two):
        return False
    return all(key in two and equal(value, two[key]) for key, value in one.items())


def _sequence_equal(one, two):
    """
    Check if two sequences are equal using the semantics of `equal`.
    """
    if len(one) != len(two):
        return False
    return all(equal(i, j) for i, j in zip(one, two))


def equal(one, two):
    """
    Check if two things are equal evading some Python type hierarchy semantics.

    Specifically in JSON Schema, evade `bool` inheriting from `int`,
    recursing into sequences to do the same.
    """
    if isinstance(one, str) or isinstance(two, str):
        try:
            return str(one) == str(two)
        except ValueError:
            return False
    if isinstance(one, Sequence) and isinstance(two, Sequence):
        return _sequence_equal(one, two)
    if isinstance(one, Mapping) and isinstance(two, Mapping):
        return _mapping_equal(one, two)
    return unbool(one) == unbool(two)


def unbool(element, true=object(), false=object()):
    """
    A hack to make True and 1 and False and 0 unique for ``uniq``.
    """

    if element is True:
        return true
    elif element is False:
        return false
    return element


def uniq(container):
    """
    Check if all of a container's elements are unique.

    Tries to rely on the container being recursively sortable, or otherwise
    falls back on (slow) brute force.
    """
    c = []
    for i in container:
        i = unbool(i)
        if isinstance(i, (int, float, bool)):
            i = str(i)
        c.append(i)
    try:
        sort = sorted(unbool(i) for i in c)
        sliced = itertools.islice(sort, 1, None)

        for i, j in zip(sort, sliced):
            if equal(i, j):
                return False

    except (NotImplementedError, TypeError):
        seen = []
        for e in c:
            e = unbool(e)
            for i in seen:
                if equal(i, e):
                    return False

            seen.append(e)
    return True


def uniq_keys(container, keys):
    """
    Check if all of a container's elements are unique based on a key.

    Tries to rely on the container being recursively sortable, or otherwise
    falls back on (slow) brute force.
    """
    c = []
    for i in container:
        try:
            o = {}
            for k in keys:
                o[k] = i[k]
            c.append(o)
        except (KeyError, TypeError):
            pass

    seen = []
    for e in c:
        e = unbool(e)
        for i in seen:
            if equal(i, e):
                return False

        seen.append(e)
    return True



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.jsonschema import CfnTemplateValidator
from cfnlint.rules.functions.GetAz import GetAz
from cfnlint.rules.functions.Ref import Ref
from cfnlint.rules.outputs.Export import Export


@pytest.fixture(scope="module")
def rule():
    rule = Export()
    yield rule


@pytest.fixture(scope="module")
def validator():
    yield CfnTemplateValidator(schema={}).extend(
        validators={
            "fn_getazs": GetAz().fn_getazs,
            "ref": Ref().ref,
        }
    )(schema={})


@pytest.mark.parametrize(
    "name,instance,errors",
    [
        ("Valid with string", "foo", []),
        ("Invalid with an array", ["foo"], ["['foo'] is not of type 'string'"]),
        (
            "Invalid with a function that returns an array",
            {"Fn::GetAZs": ""},
            ["{'Fn::GetAZs': ''} is not of type 'string'"],
        ),
        (
            "Valid with a function that returns a string",
            {"Ref": "AWS::Region"},
            [],
        ),
    ],
)
def test_functions(name, instance, errors, rule, validator):
    errs = list(rule.validate(validator, {}, instance, {}))
    assert len(errs) == len(errors), name
    for i, err in enumerate(errors):
        assert err == errs[i].message



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import UserDict
from typing import TYPE_CHECKING, Any

import regex as re

from cfnlint.helpers import ensure_list

if TYPE_CHECKING:
    from cfnlint.schema import Schema


_all_property_types = [
    "AWS::Amplify::Branch",
    "AWS::Amplify::Domain",
    "AWS::AppSync::DomainName",
    "AWS::AppSync::FunctionConfiguration",
    "AWS::AppSync::Resolver",
    "AWS::AutoScaling::AutoScalingGroup",
    "AWS::Backup::BackupSelection",
    "AWS::Backup::BackupVault",
    "AWS::CodeArtifact::Domain",
    "AWS::CodeArtifact::Repository",
    "AWS::EC2::CapacityReservation",
    "AWS::EC2::Instance",
    "AWS::EC2::SecurityGroup",
    "AWS::EC2::Subnet",
    "AWS::EC2::VPC",
    "AWS::EFS::MountTarget",
    "AWS::EKS::Nodegroup",
    "AWS::ElasticLoadBalancingV2::LoadBalancer",
    "AWS::Events::EventBus",
    "AWS::EventSchemas::Discoverer",
    "AWS::EventSchemas::Registry",
    "AWS::EventSchemas::Schema",
    "AWS::GameLift::GameSessionQueue",
    "AWS::GameLift::MatchmakingConfiguration",
    "AWS::GameLift::MatchmakingRuleSet",
    "AWS::Grafana::Workspace",
    "AWS::Greengrass::ConnectorDefinition",
    "AWS::Greengrass::CoreDefinition",
    "AWS::Greengrass::DeviceDefinition",
    "AWS::Greengrass::FunctionDefinition",
    "AWS::Greengrass::Group",
    "AWS::Greengrass::LoggerDefinition",
    "AWS::Greengrass::ResourceDefinition",
    "AWS::Greengrass::SubscriptionDefinition",
    "AWS::ImageBuilder::Component",
    "AWS::ImageBuilder::ContainerRecipe",
    "AWS::ImageBuilder::DistributionConfiguration",
    "AWS::ImageBuilder::ImagePipeline",
    "AWS::ImageBuilder::ImageRecipe",
    "AWS::ImageBuilder::InfrastructureConfiguration",
    "AWS::Neptune::DBCluster",
    "AWS::RDS::DBInstance",
    "AWS::RDS::DBParameterGroup",
    "AWS::RoboMaker::RobotApplication",
    "AWS::RoboMaker::SimulationApplication",
    "AWS::Route53Resolver::ResolverRule",
    "AWS::Route53Resolver::ResolverRuleAssociation",
    "AWS::S3::AccessPoint",
    "AWS::SageMaker::DataQualityJobDefinition",
    "AWS::SageMaker::ModelBiasJobDefinition",
    "AWS::SageMaker::ModelExplainabilityJobDefinition",
    "AWS::SageMaker::ModelQualityJobDefinition",
    "AWS::SageMaker::MonitoringSchedule",
    "AWS::SNS::Topic",
    "AWS::SQS::Queue",
    "AWS::SSM::Parameter",
    "AWS::StepFunctions::Activity",
]

# Non registry resources that have a difference between readOnlyProperties
# what is supported by GetAtt
_exceptions = {
    "AWS::AppMesh::GatewayRoute": [
        "GatewayRouteName",
        "MeshName",
        "MeshOwner",
        "VirtualGatewayName",
    ],
    "AWS::AppMesh::Mesh": [
        "MeshName",
    ],
    "AWS::AppMesh::Route": [
        "MeshName",
        "MeshOwner",
        "RouteName",
        "VirtualRouterName",
    ],
    "AWS::AppMesh::VirtualGateway": [
        "MeshName",
        "MeshOwner",
        "VirtualGatewayName",
    ],
    "AWS::AppMesh::VirtualNode": [
        "MeshName",
        "MeshOwner",
        "VirtualNodeName",
    ],
    "AWS::AppMesh::VirtualRouter": [
        "MeshName",
        "MeshOwner",
        "VirtualRouterName",
    ],
    "AWS::AppMesh::VirtualService": [
        "MeshName",
        "MeshOwner",
        "VirtualServiceName",
    ],
    "AWS::AppSync::DataSource": [
        "Name",
    ],
    "AWS::Cloud9::EnvironmentEC2": [
        "Name",
    ],
    "AWS::CloudWatch::InsightRule": [
        "RuleName",
    ],
    "AWS::DocDB::DBCluster": [
        "Port",
    ],
    "AWS::EC2::Instance": [
        "AvailabilityZone",
    ],
    "AWS::EC2::SecurityGroup": ["VpcId"],
    "AWS::Greengrass::ConnectorDefinition": [
        "Name",
    ],
    "AWS::Greengrass::CoreDefinition": [
        "Name",
    ],
    "AWS::Greengrass::DeviceDefinition": [
        "Name",
    ],
    "AWS::Greengrass::FunctionDefinition": [
        "Name",
    ],
    "AWS::Greengrass::Group": [
        "Name",
        "RoleArn",
    ],
    "AWS::Greengrass::LoggerDefinition": [
        "Name",
    ],
    "AWS::Greengrass::ResourceDefinition": [
        "Name",
    ],
    "AWS::Greengrass::SubscriptionDefinition": [
        "Name",
    ],
    "AWS::IoT1Click::Device": [
        "Enabled",
    ],
    "AWS::IoT1Click::Placement": [
        "PlacementName",
        "ProjectName",
    ],
    "AWS::IoT1Click::Project": [
        "ProjectName",
    ],
    "AWS::Kinesis::StreamConsumer": [
        "ConsumerName",
        "StreamARN",
    ],
    "AWS::ManagedBlockchain::Member": [
        "NetworkId",
    ],
    "AWS::MediaConvert::JobTemplate": [
        "Name",
    ],
    "AWS::MediaConvert::Preset": [
        "Name",
    ],
    "AWS::MediaConvert::Queue": [
        "Name",
    ],
    "AWS::MediaLive::Input": [
        "Destinations",
        "Sources",
    ],
    "AWS::OpsWorks::Instance": ["AvailabilityZone"],
    "AWS::OpsWorks::UserProfile": ["SshUsername"],
    "AWS::Route53Resolver::ResolverEndpoint": [
        "Direction",
        "Name",
    ],
    "AWS::SageMaker::CodeRepository": ["CodeRepositoryName"],
    "AWS::SageMaker::Endpoint": [
        "EndpointName",
    ],
    "AWS::SageMaker::EndpointConfig": ["EndpointConfigName"],
    "AWS::SageMaker::Model": ["ModelName"],
    "AWS::SageMaker::NotebookInstance": ["NotebookInstanceName"],
    "AWS::SageMaker::NotebookInstanceLifecycleConfig": [
        "NotebookInstanceLifecycleConfigName"
    ],
    "AWS::SageMaker::Workteam": ["WorkteamName"],
    "AWS::ServiceDiscovery::Service": ["Name"],
    "AWS::Transfer::User": [
        "ServerId",
        "UserName",
    ],
}

_unnamed_unknown_types = (
    "Custom::",
    "AWS::Serverless::",
    "AWS::CloudFormation::CustomResource",
    "Module",
)


class AttributeDict(UserDict):
    def __init__(self, __dict: None = None) -> None:
        super().__init__(__dict)
        self.data: dict[str, str] = {}

    def __getitem__(self, key: str) -> str:
        possible_items = {}
        for k, v in self.data.items():
            if re.fullmatch(k, key) or k == key:
                possible_items[k] = v
        if not possible_items:
            raise KeyError(key)
        longest_match = sorted(possible_items.keys(), key=len)[-1]
        return possible_items[longest_match]

    def get(self, key: str, default: Any = None) -> Any:
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def __repr__(self) -> str:
        keys = []
        for k in self.data.keys():
            keys.append(k.replace("\\", ""))
        return f"{keys!r}"


class GetAtts:
    """Class for helping with GetAtt logic"""

    def __init__(self, schema: "Schema") -> None:
        self._attrs = AttributeDict()
        if schema.type_name in _all_property_types:
            for name, value in schema.schema.get("properties", {}).items():
                self._process_schema(value, schema, f"/properties/{name}")
            return
        if schema.type_name in _exceptions:
            for name in _exceptions[schema.type_name]:
                self._attrs[self._pointer_to_attr(f"/properties/{name}")] = (
                    f"/properties/{name}"
                )

        if schema.type_name == "AWS::CloudFormation::Stack":
            self._attrs["Outputs\\..*"] = "/properties/CfnLintStringType"
            return

        if schema.type_name == "AWS::ServiceCatalog::CloudFormationProvisionedProduct":
            for ro_attr in schema.schema.get("readOnlyProperties", []):
                if ro_attr == "/properties/Outputs":
                    self._attrs["Outputs\\..*"] = "/properties/CfnLintStringType"
                else:
                    self._attrs[self._pointer_to_attr(ro_attr)] = ro_attr
            return

        for unnamed_type in _unnamed_unknown_types:
            if schema.type_name.startswith(unnamed_type):
                self._attrs[".*"] = "/properties/CfnLintAllTypes"

        for ro_attr in schema.schema.get("readOnlyProperties", []):
            self._attrs[self._pointer_to_attr(ro_attr)] = ro_attr

    def _pointer_to_attr(self, pointer: str) -> str:
        return ".".join(pointer.split("/")[2:])

    def _process_schema(self, obj: dict[str, Any], schema: "Schema", path: str):
        types = ensure_list(obj.get("type"))
        if "object" in types:
            for prop, value in obj.get("properties", {}).items():
                self._process_schema(value, schema, f"{path}/{prop}")
        elif "array" in types:
            # GetAtt doesn't support going into an array of objects or another array
            # so we only look at an array of strings, etc.
            if obj.get("items", {}).get("type") in ["object", "array"]:
                return
            self._attrs[self._pointer_to_attr(path)] = path
        elif obj.get("$ref"):
            _, obj = schema.resolver.resolve(obj.get("$ref"))
            self._process_schema(obj, schema, path)
        else:
            self._attrs[self._pointer_to_attr(path)] = path

    @property
    def attrs(self) -> AttributeDict:
        return self._attrs



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import unittest

from cfnlint.jsonschema._types import TypeChecker, is_null
from cfnlint.jsonschema.exceptions import UndefinedTypeCheck


class TestCfnTypeChecker(unittest.TestCase):
    def test_class(self):
        emptyChecker = TypeChecker()
        self.assertEqual(emptyChecker.type_checkers, {})
        self.assertEqual(str(emptyChecker), "<TypeChecker types={}>")

        with self.assertRaises(UndefinedTypeCheck):
            emptyChecker.is_type("foo", "bar")

    def test_null(self):
        # Only a small number of places can use null
        self.assertTrue(is_null(None, None))
        self.assertFalse(is_null(None, ""))



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_apigateway_restapi
from cfnlint.jsonschema.exceptions import ValidationError
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class RestApiOpenApi(CfnLintJsonSchema):
    id = "E3660"
    shortdesc = "RestApi requires a name when not using an OpenAPI specification"
    description = (
        "When using AWS::ApiGateway::RestApi you have to provide 'Name' "
        "if you don't provide 'Body' or 'BodyS3Location'"
    )
    tags = ["resources", "apigateway"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::ApiGateway::RestApi/Properties"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_apigateway_restapi,
                filename="openapi_properties.json",
            ),
            all_matches=False,
        )

    def message(self, instance: Any, err: ValidationError) -> str:
        return f"{err.message} when not specifying one of ['Body', 'BodyS3Location']"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_ecs_service
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class FargateDeploymentSchedulingStrategy(CfnLintJsonSchema):
    id = "E3044"
    shortdesc = (
        "ECS service using FARGATE or EXTERNAL can only "
        "use SchedulingStrategy of REPLICA"
    )
    description = (
        "When using a TargetType of Fargate or External the SchedulingStrategy "
        "has to be Replica"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ecs-service.html#cfn-ecs-service-schedulingstrategy"
    tags = ["properties", "ecs", "service", "container", "fargate"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::ECS::Service/Properties"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_ecs_service,
                filename="fargate.json",
            ),
        )

    def message(self, instance: Any, err: ValidationError) -> str:
        return err.message



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from typing import Any, Iterator

from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.rules import CloudFormationLintRule


class ImportValue(CloudFormationLintRule):
    """Check if a Output is done of another output"""

    id = "W6001"
    shortdesc = "Check Outputs using ImportValue"
    description = (
        "Check if the Output value is set using ImportValue, so creating an Output of"
        " an Output"
    )
    source_url = "https://github.com/aws-cloudformation/cfn-lint"
    tags = ["outputs", "importvalue"]

    def validate(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ) -> Iterator[ValidationError]:
        if len(validator.context.path.path) >= 2:
            if (
                validator.context.path.path[0] == "Outputs"
                and validator.context.path.path[2] == "Value"
            ):
                key = list(instance.keys())[0]
                yield ValidationError(
                    (f"The output value {instance!r} is an import from another output"),
                    rule=self,
                    path=deque([key]),
                )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import os
from test.testlib.testcase import BaseTestCase

from cfnlint.config import ConfigMixIn
from cfnlint.runner import Runner, UnexpectedRuleException


class TestGetRules(BaseTestCase):
    """Test Run Checks"""

    def test_invalid_rule(self):
        """test invalid rules"""
        err = None
        config = ConfigMixIn(
            append_rules=["invalid"],
        )

        try:
            Runner(config)
        except UnexpectedRuleException as e:
            err = e
        self.assertIsInstance(err, UnexpectedRuleException)

    def test_append_module(self):
        """test appending rules from a module"""
        config = ConfigMixIn(
            append_rules=["test.fixtures.rules.custom1"],
        )
        runner = Runner(config)
        self.assertIn("E9001", runner.rules)
        # Make sure the default rules are there too.
        self.assertIn("E1001", runner.rules)

    def test_append_directory(self):
        """test appending rules from a directory"""
        import test.fixtures.rules

        path = os.path.dirname(test.fixtures.rules.__file__)
        config = ConfigMixIn(
            append_rules=[path],
        )
        runner = Runner(config)
        self.assertIn("E9001", runner.rules)
        # Make sure the default rules are there too.
        self.assertIn("E1001", runner.rules)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import json
import logging
import os
import subprocess

import cfnlint
import cfnlint.data.AdditionalSpecs
from cfnlint.helpers import get_url_content
from cfnlint.schema import PROVIDER_SCHEMA_MANAGER

LOGGER = logging.getLogger(__name__)

REGISTRY_SCHEMA_ZIP = (
    "https://schema.cloudformation.us-east-1.amazonaws.com/CloudformationSchema.zip"
)


def update_resource_specs(force: bool = False):
    # Pool() uses cpu count if no number of processors is specified
    # Pool() only implements the Context Manager protocol from Python3.3 onwards,
    # so it will fail Python2.7 style linting, as well as throw AttributeError

    # Update provider Schemas
    PROVIDER_SCHEMA_MANAGER.update(force)


def update_documentation(rules):
    # Update the overview of all rules in the linter
    filename = "docs/rules.md"

    # Sort rules by the Rule ID
    sorted_rules = sorted(rules.values(), key=lambda obj: obj.id)
    data = []

    # Read current file up to the Rules part, everything up to that point is
    # static documentation.
    with open(filename, "r", encoding="utf-8") as original_file:
        line = original_file.readline()
        while line:
            data.append(line)

            if line == "## Rules\n":
                break

            line = original_file.readline()

    # Rebuild the file content
    with open(filename, "w", encoding="utf-8") as new_file:
        # Rewrite the static documentation
        for line in data:
            new_file.write(line)

        # Add the rules
        new_file.write(
            "(_This documentation is generated by running `cfn-lint"
            " --update-documentation`, do not alter this manually_)\n\n"
        )
        new_file.write(
            f"The following **{len(sorted_rules) + 3}** rules are applied by this"
            " linter:\n\n"
        )
        new_file.write(
            "| Rule ID  | Title | Description | Config<br />(Name:Type:Default) |"
            " Source | Tags |\n"
        )
        new_file.write(
            "| -------- | ----- | ----------- | ---------- | ------ | ---- |\n"
        )

        rule_output = (
            '| [{0}<a name="{0}"></a>]({6}) | {1} | {2} | {3} | [Source]({4}) | {5} |\n'
        )

        for rule in sorted_rules:
            rule_source_code_file = (
                "../"
                + subprocess.check_output(
                    [
                        "git",
                        "grep",
                        "-l",
                        'id = "' + rule.id + '"',
                        "src/cfnlint/rules/",
                    ]
                )
                .decode("ascii")
                .strip()
            )
            rule_id = rule.id + "*" if rule.experimental else rule.id
            tags = ",".join(f"`{tag}`" for tag in rule.tags)
            config = "<br />".join(
                f'{key}:{values.get("type")}:{values.get("default")}'
                for key, values in rule.config_definition.items()
            )
            new_file.write(
                rule_output.format(
                    rule_id,
                    rule.shortdesc,
                    rule.description,
                    config,
                    rule.source_url,
                    tags,
                    rule_source_code_file,
                )
            )
        new_file.write("\n\\* experimental rules\n")


def update_iam_policies():
    """update iam policies file"""

    url = "https://awspolicygen.s3.amazonaws.com/js/policies.js"

    filename = os.path.join(
        os.path.dirname(cfnlint.data.AdditionalSpecs.__file__), "Policies.json"
    )
    LOGGER.debug("Downloading policies %s into %s", url, filename)

    content = get_url_content(url)

    content = content.split("app.PolicyEditorConfig=")[1]
    content = json.loads(content)

    actions = {
        "Manage Amazon API Gateway": ["HEAD", "OPTIONS"],
        "Amazon API Gateway Management": ["HEAD", "OPTIONS"],
        "Amazon API Gateway Management V2": ["HEAD", "OPTIONS"],
        "Amazon Kinesis Video Streams": ["StartStreamEncryption"],
    }
    for k, v in actions.items():
        if content.get("serviceMap").get(k):
            content["serviceMap"][k]["Actions"].extend(v)
        else:
            LOGGER.debug('"%s" was not found in the policies file', k)

    with open(filename, "w", encoding="utf-8") as f:
        json.dump(content, f, indent=1, sort_keys=True, separators=(",", ": "))
        f.write("\n")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from typing import Any

from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.rules import CloudFormationLintRule


class NoEcho(CloudFormationLintRule):
    id = "W2010"
    shortdesc = "NoEcho parameters are not masked when used in Metadata and Outputs"
    description = (
        "Using the NoEcho attribute does not mask any information stored "
        "in the following: Metadata, Outputs, Resource Metadata"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/parameters-section-structure.html"
    tags = ["functions", "dynamic reference", "ref"]

    def validate(self, validator: Validator, _, instance: Any, schema: Any):
        value = instance.get("Ref")

        if not validator.is_type(value, "string"):
            return

        parameter = validator.context.parameters.get(value)
        if not parameter:
            return

        if parameter.no_echo:
            if len(validator.context.path.path) >= 3:
                if (
                    validator.context.path.path[0] == "Resources"
                    and validator.context.path.path[2] == "Metadata"
                ):
                    yield ValidationError(
                        f"Don't use 'NoEcho' parameter {value!r} in resource metadata",
                        rule=self,
                        path=deque(["Ref"]),
                    )
                    return
            if len(validator.context.path.path) > 0:
                if validator.context.path.path[0] in ["Metadata", "Outputs"]:
                    yield ValidationError(
                        (
                            f"Don't use 'NoEcho' parameter {value!r} "
                            f"in {validator.context.path.path[0]!r}"
                        ),
                        rule=self,
                        path=deque(["Ref"]),
                    )
                    return



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.lmbd.PermissionSourceAccount import PermissionSourceAccount


@pytest.fixture(scope="module")
def rule():
    rule = PermissionSourceAccount()
    yield rule


@pytest.fixture
def template():
    return {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Conditions": {
            "IsUsEast1": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]},
        },
        "Resources": {
            "Bucket": {"Type": "AWS::S3::Bucket"},
            "SQS": {"Type": "AWS::SQS::Queue"},
        },
    }


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "SourceArn": "arn:aws:s3:::bucket_name",
                "SourceAccount": "123456789012",
            },
            [],
        ),
        (
            {},
            [],
        ),
        (
            [],
            [],
        ),
        (
            {
                "SourceArn": [],
            },
            [],
        ),
        (
            {
                "SourceArn": "arn:aws:s3:::bucket_name",
            },
            [
                ValidationError(
                    "'SourceAccount' is a required property",
                    validator="required",
                )
            ],
        ),
        (
            {
                "SourceArn": "arn:aws:sqs:us-east-1:123456789012:queue",
            },
            [],
        ),
        (
            {
                "SourceArn": {
                    "Fn::Sub": (
                        "arn:${AWS::Partition}:sqs:"
                        "${AWS::Region}:${AWS::AccountId}:queue"
                    )
                },
            },
            [],
        ),
        (
            {
                "SourceArn": {"Fn::Sub": "arn:${AWS::Partition}:s3:::bucket"},
            },
            [
                ValidationError(
                    "'SourceAccount' is a required property",
                    validator="required",
                )
            ],
        ),
        (
            {
                "SourceArn": {"Fn::Sub": [[], {}]},
            },
            [],
        ),
        (
            {
                "SourceArn": {"Fn::Sub": "arn:${AWS::Partition}:s3:::bucket"},
                "SourceAccount": {"Ref": "AWS::AccountId"},
            },
            [],
        ),
        (
            {
                "SourceArn": {"Fn::GetAtt": ["Bucket", "Arn"]},
                "SourceAccount": {"Ref": "AWS::AccountId"},
            },
            [],
        ),
        (
            {
                "SourceArn": {"Fn::GetAtt": ["Bucket", "Arn"]},
            },
            [
                ValidationError(
                    "'SourceAccount' is a required property",
                    validator="required",
                )
            ],
        ),
        (
            {
                "SourceArn": {"Fn::GetAtt": ["SQS", "Arn"]},
                "SourceAccount": {"Ref": "AWS::AccountId"},
            },
            [],
        ),
        (
            {
                "SourceArn": {"Fn::GetAtt": ["SQS", "Arn"]},
            },
            [],
        ),
        (
            {
                "SourceArn": {"Ref": "Foo"},
            },
            [],
        ),
        (
            {
                "SourceArn": {
                    "Fn::If": [
                        "IsUsEast1",
                        {"Fn::GetAtt": ["Bucket", "Arn"]},
                        {"Fn::GetAtt": ["SQS", "Arn"]},
                    ]
                },
                "SourceAccount": {
                    "Fn::If": [
                        "IsUsEast1",
                        {"Ref": "AWS::AccountId"},
                        {"Ref": "AWS::NoValue"},
                    ]
                },
            },
            [],
        ),
        (
            {
                "SourceArn": {
                    "Fn::If": [
                        "IsUsEast1",
                        {"Fn::GetAtt": ["Bucket", "Arn"]},
                        {"Fn::GetAtt": ["SQS", "Arn"]},
                    ]
                },
                "SourceAccount": {
                    "Fn::If": [
                        "IsUsEast1",
                        {"Ref": "AWS::NoValue"},
                        {"Ref": "AWS::AccountId"},
                    ]
                },
            },
            [
                ValidationError(
                    "'SourceAccount' is a required property",
                    validator="required",
                )
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError, _keywords
from cfnlint.jsonschema.validators import CfnTemplateValidator
from cfnlint.rules import CloudFormationLintRule


class Error(CloudFormationLintRule):
    id = "E1111"

    def validate(self, validator, s, instance, schema):
        if s:
            yield ValidationError(
                "Error",
                rule=self,
            )


class Warning(CloudFormationLintRule):
    id = "W1111"

    def validate(self, validator, s, instance, schema):
        yield ValidationError(
            "Warning",
            rule=self,
        )


@pytest.fixture
def validator():
    validator = CfnTemplateValidator(schema={})
    validator = validator.extend(
        validators={
            "error": Error().validate,
            "warning": Warning().validate,
        }
    )
    return validator({})


@pytest.mark.parametrize(
    "name,instance,schema,expected",
    [
        (
            "Valid anyOf",
            "foo",
            [{"const": "foo"}, {"const": "bar"}],
            [],
        ),
        (
            "Valid anyOf with error rule",
            "foo",
            [{"const": "foo"}, {"error": True}],
            [],
        ),
        (
            "Invalid anyOf with error rule",
            "foo",
            [{"error": True}, {"error": True}],
            [
                ValidationError(
                    "'foo' is not valid under any of the given schemas",
                    path=deque([]),
                    schema_path=deque([]),
                    context=[
                        ValidationError(
                            "Error",
                            rule=Error(),
                            path=deque([]),
                            validator="error",
                            schema_path=deque([0, "error"]),
                        ),
                        ValidationError(
                            "Error",
                            rule=Error(),
                            path=deque([]),
                            validator="error",
                            schema_path=deque([1, "error"]),
                        ),
                    ],
                ),
            ],
        ),
        (
            "Valid anyOf with a warning validation error",
            "foo",
            [{"warning": True}, {"error": True}],
            [
                ValidationError(
                    "Warning",
                    rule=Warning(),
                    path=deque([]),
                    validator="warning",
                    schema_path=deque([0, "warning"]),
                ),
            ],
        ),
        (
            "Valid anyOf with a warning validation error",
            "foo",
            [{"error": True}, {"warning": True}],
            [
                ValidationError(
                    "Warning",
                    rule=Warning(),
                    path=deque([]),
                    validator="warning",
                    schema_path=deque([1, "warning"]),
                ),
            ],
        ),
    ],
)
def test_anyof(name, instance, schema, validator, expected):
    errs = list(_keywords.anyOf(validator, schema, instance, schema))
    assert errs == expected, f"{name!r} got errors {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.lmbd.SnapStartEnabled import SnapStartEnabled


@pytest.fixture(scope="module")
def rule():
    rule = SnapStartEnabled()
    yield rule


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "python doesn't need SnapStart",
            "python3.11",
            [],
        ),
        (
            "wrong type",
            [],
            [],
        ),
        (
            "java 8 doesn't need SnapStart",
            "java8",
            [],
        ),
        (
            "java 17 should have SnapStart",
            "java17",
            [
                ValidationError(
                    "'java17' runtime should consider using 'SnapStart'",
                    path=deque(["SnapStart", "ApplyOn"]),
                    rule=SnapStartEnabled(),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, expected, rule):
    errs = list(rule.validate(instance))

    assert errs == expected, f"{name!r}: expected {expected!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_rds_dbinstance
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class DbInstanceAuroraExclusive(CfnLintJsonSchema):
    id = "E3682"
    shortdesc = "Validate when using Aurora certain properies aren't required"
    description = (
        "When creating an aurora DBInstance don't specify "
        "'AllocatedStorage', 'BackupRetentionPeriod', 'CopyTagsToSnapshot', "
        "'DeletionProtection', 'EnableIAMDatabaseAuthentication', "
        "'MasterUserPassword', or 'StorageEncrypted'"
    )
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::RDS::DBInstance/Properties"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_rds_dbinstance,
                filename="aurora_exclusive.json",
            ),
        )

    def message(self, instance: Any, err: ValidationError) -> str:
        keys = [
            "AllocatedStorage",
            "BackupRetentionPeriod",
            "CopyTagsToSnapshot",
            "DeletionProtection",
            "EnableIAMDatabaseAuthentication",
            "MasterUserPassword",
            "StorageEncrypted",
        ]
        extra = []
        for property in instance.keys():
            if property in keys:
                extra.append(property)

        return f"Additional properties are not allowed ({extra!r})"

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        if not validator.is_type(instance, "object"):
            return

        if validator.is_type(instance.get("Engine"), "string"):
            instance["Engine"] = instance["Engine"].lower()

        yield from super().validate(validator, keywords, instance, schema)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from test.unit.rules import BaseRuleTestCase
from unittest.mock import Mock, patch

from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.resources.ResourceType import ResourceType


class TestType(BaseRuleTestCase):
    """Test AWS Types"""

    def setUp(self):
        """Setup"""
        super(TestType, self).setUp()
        self.rule = ResourceType()

    def test_property_names(self):
        validator = CfnTemplateValidator({})
        errors = list(
            self.rule.validate(
                validator, "cfnResourceType", {"Type": "AWS::S3::Bucket"}, {}
            )
        )
        self.assertListEqual(
            errors,
            [],
            errors,
        )

    @patch("cfnlint.template.Template", autospec=True)
    def test_types_with_conditions(self, cfn):
        cfn = Mock()
        cfn.conditions = Mock()
        cfn.conditions.build_scenerios_on_region.return_value = [True]
        validator = CfnTemplateValidator({}).evolve(cfn=cfn)
        errors = list(
            self.rule.validate(
                validator,
                "cfnResources",
                {"Type": "Foo::Bar::Type", "Condition": "IsUsEast1"},
                {},
            )
        )

        self.assertListEqual(
            errors,
            [
                ValidationError(
                    "Resource type 'Foo::Bar::Type' does not exist in 'us-east-1'",
                    path=deque(["Type"]),
                    schema_path=deque([]),
                    rule=ResourceType(),
                ),
            ],
            errors,
        )

    @patch("cfnlint.template.Template", autospec=True)
    def test_types_with_conditions_false(self, cfn):
        cfn = Mock()
        cfn.conditions = Mock()
        cfn.conditions.build_scenerios_on_region.return_value = [False]
        validator = CfnTemplateValidator({}).evolve(cfn=cfn)
        errors = list(
            self.rule.validate(
                validator,
                "cfnResources",
                {"Type": "Foo::Bar::Type", "Condition": "IsUsEast1"},
                {},
            )
        )

        self.assertListEqual(errors, [], errors)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint import ConfigMixIn
from cfnlint.rules.jsonschema.JsonSchema import JsonSchema


class TestBaseTemplate(BaseRuleTestCase):
    """Test base template"""

    def setUp(self):
        """Setup"""
        super(TestBaseTemplate, self).setUp()
        self.collection.register(JsonSchema())

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_file_positive_configured(self):
        self.helper_file_positive_template(
            "test/fixtures/templates/bad/generic.yaml",
            ConfigMixIn(
                [],
                configure_rules={
                    "E1001": {
                        "sections": "Errors",
                    }
                },
            ),
        )

    def test_file_negative(self):
        """Test failure"""
        self.helper_file_negative("test/fixtures/templates/bad/generic.yaml", 1)

    def test_file_base(self):
        """Test failure"""
        self.helper_file_negative("test/fixtures/templates/bad/templates/base.yaml", 1)

    def test_file_base_date(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/templates/base_date.yaml", 1
        )

    def test_file_base_null(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/templates/base_null.yaml", 2
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from unittest import TestCase

from cfnlint.jsonschema import CfnTemplateValidator
from cfnlint.rules.resources.iam.ResourceEcrPolicy import ResourceEcrPolicy


class TestResourceEcrPolicy(TestCase):
    """Test IAM identity Policies"""

    def setUp(self):
        """Setup"""
        self.rule = ResourceEcrPolicy()

    def test_string_statements(self):
        """Test Positive"""
        validator = CfnTemplateValidator()

        # no resource statement
        # ruff: noqa: E501
        policy = """
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Sid": "AllowCrossAccountPush",
                        "Effect": "Allow",
                        "Principal": {
                            "AWS": "arn:aws:iam::123456789012:root"
                        },
                        "Action": [
                            "ecr:BatchCheckLayerAvailability",
                            "ecr:CompleteLayerUpload",
                            "ecr:InitiateLayerUpload",
                            "ecr:PutImage",
                            "ecr:UploadLayerPart"
                        ]
                    }
                ]
            }
        """

        errs = list(
            self.rule.validate(
                validator=validator, policy=policy, schema={}, policy_type=None
            )
        )
        self.assertEqual(len(errs), 0, errs)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import cfnlint.data.schemas.extensions.aws_docdb_dbinstance
from cfnlint.rules.jsonschema.CfnLintJsonSchema import SchemaDetails
from cfnlint.rules.jsonschema.CfnLintJsonSchemaRegional import CfnLintJsonSchemaRegional


class DBInstanceClassEnum(CfnLintJsonSchemaRegional):
    id = "E3620"
    shortdesc = "Validate a DocDB DB Instance class"
    description = (
        "Validates the DocDB instance types based on region "
        "and data gathered from the pricing APIs"
    )
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::DocDB::DBInstance/Properties/DBInstanceClass"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_docdb_dbinstance,
                filename="dbinstanceclass_enum.json",
            ),
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.testlib.testcase import BaseTestCase
from typing import List

from cfnlint import ConfigMixIn, Rules
from cfnlint.config import ManualArgs
from cfnlint.runner import TemplateRunner


class BaseRuleTestCase(BaseTestCase):
    """Used for Testing Rules"""

    success_templates: List[str] = []

    def setUp(self):
        """Setup"""
        self.collection = Rules()
        self.config = ConfigMixIn(
            include_experimental=True,
            include_checks=["I"],
        )

    def helper_file_positive(self, config=None):
        """Success test"""
        config = config or self.config
        for filename in self.success_templates:
            template = self.load_template(filename)
            good_runner = TemplateRunner(filename, template, config, self.collection)
            failures = list(good_runner.run())
            self.assertEqual(
                [], failures, "Got failures {} on {}".format(failures, filename)
            )

    def helper_file_positive_template(self, filename, config=None):
        """Success test with template parameter"""
        config = config or self.config
        template = self.load_template(filename)
        good_runner = TemplateRunner(filename, template, config, self.collection)
        failures = list(good_runner.run())
        self.assertEqual(
            [],
            failures,
            "Expected {} failures but got {} on {}".format(0, failures, filename),
        )

    def helper_file_negative(self, filename, err_count, config=None):
        """Failure test"""
        config = config or self.config
        template = self.load_template(filename)
        bad_runner = TemplateRunner(filename, template, config, self.collection)
        failures = list(bad_runner.run())
        self.assertEqual(
            err_count,
            len(failures),
            "Expected {} failures but got {} on {}".format(
                err_count, failures, filename
            ),
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import cfnlint.data.schemas.extensions.aws_emr_cluster
from cfnlint.rules.jsonschema.CfnLintJsonSchema import SchemaDetails
from cfnlint.rules.jsonschema.CfnLintJsonSchemaRegional import CfnLintJsonSchemaRegional


class ClusterInstanceTypeConfigInstanceTypeEnum(CfnLintJsonSchemaRegional):
    id = "E3675"
    shortdesc = "Validate EMR cluster instance type"
    description = (
        "Validates the EMR cluster instance types based on region "
        "and data gathered from the pricing APIs"
    )
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::EMR::Cluster/Properties/Instances/CoreInstanceFleet/InstanceTypeConfigs/*/InstanceType",
                "Resources/AWS::EMR::Cluster/Properties/Instances/CoreInstanceGroup/InstanceType",
                "Resources/AWS::EMR::Cluster/Properties/Instances/TaskInstanceFleets/*/InstanceTypeConfigs/*/InstanceType",
                "Resources/AWS::EMR::Cluster/Properties/Instances/TaskInstanceGroups/*/InstanceType",
            ],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_emr_cluster,
                filename="instancetypeconfig_instancetype_enum.json",
            ),
        )



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

__all__ = ["Conditions"]

from cfnlint.conditions._errors import UnknownSatisfisfaction
from cfnlint.conditions.conditions import Conditions
